from __future__ import annotations

from prettyqt import constants, core


class FlattenTreeProxyModel(core.AbstractProxyModel):
    """Proxy model to flatten a tree to appear like a table.

    ### Example

    Original model:

    === "Without proxy"

        ```py
        table = widgets.TreeView()
        source_model = itemmodels.ParentClassTreeModel(widgets.Frame)
        table.set_model(source_model)
        # table.proxifier.flatten()
        ```
        <figure markdown>
          ![Image title](../../images/flattentreeproxymodel_without.png)
        </figure>

    === "With proxy"

        ```py
        table = widgets.TreeView()
        source_model = itemmodels.ParentClassTreeModel(widgets.Frame)
        table.set_model(source_model)
        table.proxifier.flatten()
        ```
        <figure markdown>
          ![Image title](../../images/flattentreeproxymodel.png)
        </figure>

    === "Path mode"

        ```py
        table = widgets.TreeView()
        source_model = itemmodels.ParentClassTreeModel(widgets.Frame)
        table.set_model(source_model)
        table.proxifier.flatten(show_path=True)
        ```
        <figure markdown>
          ![Image title](../../images/flattentreeproxymodel_path.png)
        </figure>

    === "Leaves only mode"

        ```py
        table = widgets.TreeView()
        source_model = itemmodels.ParentClassTreeModel(widgets.Frame)
        table.set_model(source_model)
        table.proxifier.flatten(leaves_only=True)
        ```
        <figure markdown>
          ![Image title](../../images/flattentreeproxymodel_leaves_only.png)
        </figure>

    ```py
    table.proxifier.flatten()
    # or
    proxy = itemmodels.FlattenTreeProxyModel()
    proxy.set_source_model(model)
    table.set_model(proxy)
    ```
    """

    ID = "flatten_tree"

    def __init__(self, parent: widgets.QWidget | None = None, **kwargs):
        self._leaves_only = False
        self._show_path = False
        self._source_column = 0
        self.PATH_SEPARATOR = " / "
        self._source_root_index = core.ModelIndex()
        self._source_key: list[tuple[int, ...]] = []
        self._source_offset: dict[tuple[int, ...], int] = {}
        super().__init__(parent, **kwargs)

    def setSourceModel(self, model: core.QAbstractItemModel):
        if (old_model := self.sourceModel()) is not None:
            old_model.dataChanged.disconnect(self._source_data_changed)
            old_model.rowsInserted.disconnect(self._on_reset)
            old_model.rowsRemoved.disconnect(self._on_reset)
            old_model.rowsMoved.disconnect(self._on_row_move)
        with self.reset_model():
            super().setSourceModel(model)
            self._update_mapping()

        model.dataChanged.connect(self._source_data_changed)
        model.rowsInserted.connect(self._on_reset)
        model.rowsRemoved.connect(self._on_reset)
        model.rowsMoved.connect(self._on_row_move)

    def set_source_column(self, column: int):
        with self.reset_model():
            self._source_column = column
            self._update_mapping()

    def get_source_column(self) -> int:
        return self._source_column

    def set_root_index(self, root_index: core.ModelIndex):
        with self.reset_model():
            self._source_root_index = root_index
            self._update_mapping()

    def get_root_index(self) -> core.ModelIndex:
        return self._source_root_index

    def set_leaves_only(self, leaves_only: bool):
        if leaves_only != self._leaves_only:
            with self.reset_model():
                self._leaves_only = leaves_only
                self._update_mapping()

    def is_leaves_only(self) -> bool:
        return self._leaves_only

    def set_show_path(self, show: bool):
        if show != self._show_path:
            with self.reset_model():
                self._show_path = show

    def is_path_shown(self) -> bool:
        return self._show_path

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        if not source_index.isValid():
            return source_index
        key = self.get_index_key(source_index)
        row = self._source_offset[key] + source_index.row()
        return self.index(row, 0)

    def mapToSource(self, index: core.ModelIndex) -> core.ModelIndex:
        if not index.isValid():
            return index
        row = index.row()
        source_key_path = self._source_key[row]
        return self.source_index_from_key(source_key_path)

    def index(
        self, row: int, column: int = 0, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        return (
            core.ModelIndex()
            if parent.isValid()
            else self.createIndex(row, column, row)  # object=row)
        )

    def parent(self, child=None) -> core.ModelIndex:
        return super().parent() if child is None else core.ModelIndex()

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return 0 if parent.isValid() else len(self._source_key)

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return 0 if parent.isValid() else 1

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        flags = super().flags(index)
        return flags
        # this would disable non-leave items
        # index = self.mapToSource(index)
        # model = self.sourceModel()
        # enabled = flags & constants.ItemFlag.ItemIsEnabled
        # if model is not None and model.rowCount(index) > 0 and enabled:
        #     flags ^= constants.ItemFlag.ItemIsEnabled
        # return flags

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if role == constants.DISPLAY_ROLE and self._show_path:
            index = self.mapToSource(index)
            model = self.sourceModel()
            path = model.get_breadcrumbs_path(index)
            return self.PATH_SEPARATOR.join(str(i) for i in path)
        return super().data(index, role)

    def _update_mapping(self):
        if self.sourceModel() is None:
            return
        self._source_key, self._source_offset = self.get_source_mapping(self._leaves_only)

    def _source_data_changed(self, top: core.ModelIndex, bottom: core.ModelIndex):
        changed_indexes = [top.sibling(i, 0) for i in range(top.row(), bottom.row() + 1)]
        for ind in changed_indexes:
            self.dataChanged.emit(ind, ind)

    def _on_reset(self, parent: core.ModelIndex, start: int, end: int):
        with self.reset_model():
            self._update_mapping()

    def _on_row_move(
        self, source_parent, source_start, source_end, dest_parent, dest_row
    ):
        with self.reset_model():
            self._update_mapping()

    leaves_only = core.Property(
        bool,
        is_leaves_only,
        set_leaves_only,
        doc="Whether to only show the tree leaves",
    )
    show_path = core.Property(
        bool,
        is_path_shown,
        set_show_path,
        doc="Show the complete path in first column",
    )


if __name__ == "__main__":
    from prettyqt import itemmodels, widgets

    app = widgets.app()
    table = widgets.TreeView()
    source_model = itemmodels.ParentClassTreeModel(widgets.Frame)
    table.set_model(source_model)
    table.expandAll()
    table.proxifier.flatten(leaves_only=True)
    table.show()
    table.resize(600, 500)
    table.h_header.resize_sections("stretch")
    table.set_icon("mdi6.table-pivot")
    table.set_title("Flatten")
    app.exec()



from __future__ import annotations

import logging

from prettyqt import constants, widgets


logger = logging.getLogger(__name__)


class TableViewMixin(widgets.AbstractItemViewMixin):
    def __init__(
        self,
        *args,
        alternating_row_colors: bool = True,
        word_wrap: bool = False,
        **kwargs,
    ):
        super().__init__(
            *args,
            alternating_row_colors=alternating_row_colors,
            word_wrap=word_wrap,
            **kwargs,
        )
        class_name = type(self).__name__
        self.setObjectName(class_name)
        self.setHorizontalHeader(widgets.HeaderView("horizontal", parent=self))
        self.setVerticalHeader(widgets.HeaderView("vertical", parent=self))
        self.setAlternatingRowColors(True)
        self.setWordWrap(False)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"gridStyle": constants.PEN_STYLE}
        return maps

    @property
    def h_header(self) -> widgets.HeaderView:
        return self.horizontalHeader()

    @h_header.setter
    def h_header(self, header: widgets.QHeaderView):
        self.setHorizontalHeader(header)

    @property
    def v_header(self) -> widgets.HeaderView:
        return self.verticalHeader()

    @v_header.setter
    def v_header(self, header: widgets.QHeaderView):
        self.setVerticalHeader(header)

    def is_cell_visible(self, row: int, column: int) -> bool:
        is_in_horizontal = self.h_header.is_in_visual_range(column)
        is_in_vertical = self.v_header.is_in_visual_range(row)
        return is_in_horizontal and is_in_vertical

    def set_sorting_enabled(self, enabled: bool, do_sort: bool = False):
        """Hack to avoid direct sort when setting sorting enabled."""
        model = self.model()
        if not do_sort and model is not None:
            backup = model.sort
            model.sort = lambda x, y: None
        self.setSortingEnabled(enabled)
        if not do_sort and model is not None:
            model.sort = backup

    def setup_list_style(self):
        self.set_selection_behavior("rows")
        self.h_header.setStretchLastSection(True)
        self.v_header.set_resize_mode("fixed")
        self.v_header.set_default_section_size(28)

    def adapt_sizes(self):
        model = self.model()
        if model is not None and (model.rowCount() * model.columnCount()) < 1000:
            self.resizeColumnsToContents()
        else:
            self.h_header.resize_sections("interactive")

    def sort_by_column(self, column: int | None, ascending: bool = True):
        column = -1 if column is None else column
        order = constants.ASCENDING if ascending else constants.DESCENDING
        self.sortByColumn(column, order)

    def set_grid_style(self, style: constants.PenStyleStr | constants.PenStyle):
        """Set grid style.

        Args:
            style: grid style to use
        """
        self.setGridStyle(constants.PEN_STYLE.get_enum_value(style))

    def get_grid_style(self) -> constants.PenStyleStr:
        """Return grid style.

        Returns:
            grid style
        """
        return constants.PEN_STYLE.inverse[self.gridStyle()]

    def get_visible_section_span(
        self,
        orientation: constants.OrientationStr | constants.Orientation,
        margin: int = 0,
    ) -> tuple[int, int]:
        """Get a tuple containing the visible start/end indexes.

        if there are no items visible, return -1, -1
        """
        rect = self.viewport().rect()
        if orientation in ["horizontal", constants.HORIZONTAL]:
            start = self.columnAt(rect.left())
            count = self.model().columnCount()
            end = self.columnAt(rect.right())
        else:
            start = self.rowAt(rect.top())
            count = self.model().rowCount()
            end = self.rowAt(rect.bottom())
        if count == 0:
            return (-1, -1)
        start = max(0, start - margin)
        end = count if end == -1 else min(end + margin, count)
        return (start, end)

    def resizeColumnsToContents(self, max_columns: int | None = 500):
        colcount = self.model().columnCount()
        if max_columns is None or colcount > max_columns:
            super().resizeColumnsToContents()
            return
        to_check = min(colcount, max_columns)
        for i in range(to_check):
            self.resizeColumnToContents(i)

    def resize_visible_columns_to_contents(self, margin: int = 0):
        if not self.isVisible():
            logger.warning("trying resize_visible_columns_to_contents while not visible.")
        colcount = self.model().columnCount()
        autosized_cols = set()
        col, end = self.get_visible_section_span("horizontal", margin=margin)
        width = self.viewport().width()
        while col <= end:
            if col not in autosized_cols:
                autosized_cols.add(col)
                self.resizeColumnToContents(col)
            col += 1
            #  end may change during resize
            end = self.columnAt(width)
            end = colcount if end == -1 else end

    def auto_span(
        self,
        orientation: constants.OrientationStr
        | constants.Orientation = constants.HORIZONTAL,
        role=constants.DISPLAY_ROLE,
        start: tuple[int, int] = (0, 0),
        end: tuple[int, int] | None = None,
    ) -> list[tuple[int, int, int, int]]:
        """Set spans in given direction based on same content in given role."""
        is_horizontal = orientation in [constants.HORIZONTAL, "horizontal"]
        model = self.model()
        spans = []
        # figure out the ranges
        if is_horizontal:
            start_level = max(0, start[0])
            start_section = max(0, start[1])
            end_level = model.rowCount()
            end_section = model.columnCount()
            if end:
                end_level = min(max(0, end[0]), end_level)
                end_section = min(max(0, end[1]), end_section)
        else:
            start_level = max(0, start[1])
            start_section = max(0, start[0])
            end_level = model.columnCount()
            end_section = model.rowCount()
            if end:
                end_level = min(max(0, end[1]), end_level)
                end_section = min(max(0, end[0]), end_section)

        # adjust the spans.
        for level in range(start_level, end_level):
            match_start = None
            if is_horizontal:
                arr = [
                    model.index(level, i).data(role)
                    for i in range(start_section, end_section + 1)
                ]
                logger.debug(f"{type(self).__name__}: spanning horizontal {arr}")
            else:
                arr = [
                    model.index(i, level).data(role)
                    for i in range(start_section, end_section + 1)
                ]
                logger.debug(f"{type(self).__name__}: spanning vertical {arr}")
            for section in range(1, len(arr)):
                if arr[section] == arr[section - 1]:
                    if match_start is None:
                        match_start = section - 1
                    # If this is the last cell, need to end it
                    if section == end_section - start_section:
                        match_end = section
                        span_size = match_end - match_start + 1
                        begin = match_start + start_section
                        if is_horizontal:
                            self.setSpan(level, begin, 1, span_size)
                            spans.append((level, begin, 1, span_size))
                        else:
                            spans.append((begin, level, span_size, 1))
                            self.setSpan(begin, level, span_size, 1)
                elif match_start is not None:
                    match_end = section - 1
                    span_size = match_end - match_start + 1
                    begin = match_start + start_section
                    if is_horizontal:
                        self.setSpan(level, begin, 1, span_size)
                        spans.append((level, begin, 1, span_size))
                    else:
                        self.setSpan(begin, level, span_size, 1)
                        spans.append((begin, level, span_size, 1))
                    match_start = None
        return spans


class TableView(TableViewMixin, widgets.QTableView):
    """Default model/view implementation of a table view."""


if __name__ == "__main__":
    from prettyqt import debugging

    app = widgets.app()
    widget = debugging.example_table()
    widget.show()
    widget.resize(500, 500)
    widget.resize_visible_columns_to_contents()
    app.exec()



from __future__ import annotations

from collections.abc import Callable
import dataclasses
import logging

from typing import Any

from prettyqt import constants, core
from prettyqt.utils import fuzzy


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Filter:
    column: int
    value: Any
    role: constants.ItemDataRole


class MultiColumnFilterProxyModel(core.SortFilterProxyModel):
    """A SortFilterProxyModel which filters based on multiple columns in one go.

    This proxy uses seperate search term / values for each column in order to filter
    the source model, thus avoiding to layer proxy models in case you want to filter
    based on several columns. That way it is less demanding since filtering
    for all columns is done in one go.

    This model is used by the [FilterHeader](filterheader.md) widget in order to filter
    a table in one go, avoiding the need to layer multiple proxy models.


    ### Example

    ```py
    proxy = itemmodels.MultiColumnFilterProxyModel()
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "multi_column_filter"

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._filters: dict[int, Filter] = {}
        self.setRecursiveFilteringEnabled(True)

    def clear_filters(self):
        self._filters = {}
        self.invalidateRowsFilter()

    def set_filter_value(
        self,
        column: int,
        value: str,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if value == "" or value is None:  # False is a valid value.
            del self._filters[column]
        else:
            self._filters[column] = Filter(column=column, value=value, role=role)
        self.invalidateRowsFilter()

    def setFilterKeyColumn(self, column: int):
        raise NotImplementedError("Not supported.")

    def setFilterRole(self, column: int):
        raise NotImplementedError("Not supported.")

    def filterAcceptsRow(self, row, parent):
        source = self.sourceModel()
        for k, v in self._filters.items():
            index = source.index(row, k, parent)
            data = source.data(index, v.role)
            search_val = v.value
            match search_val:
                case str():
                    search_val = str(search_val)
                    data = str(data)
                    if self.filterMode == "fuzzy" and not fuzzy.fuzzy_match_simple(
                        search_val,
                        data,
                        case_sensitive=self.is_filter_case_sensitive(),
                    ):
                        return False
                    if not self.is_filter_case_sensitive():
                        search_val = search_val.lower()
                        data = data.lower()
                    if not data.startswith(search_val):
                        return False
                case Callable():
                    return search_val(data)
                case _:
                    if data != search_val:
                        return False
        return True


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()
    table = debugging.example_tree()
    model = MultiColumnFilterProxyModel()
    model.setSourceModel(table.model())
    table.setRootIsDecorated(True)
    table.set_model(model)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtTest


class AbstractItemModelTester(core.ObjectMixin, QtTest.QAbstractItemModelTester):
    pass


if __name__ == "__main__":
    from prettyqt import gui

    model = gui.StandardItemModel()
    tester = AbstractItemModelTester(model)



from __future__ import annotations

from typing import Literal

from prettyqt import positioning
from prettyqt.qt import QtLocation
from prettyqt.utils import bidict


InstructionDirectionStr = Literal[
    "none",
    "forward",
    "bear_right",
    "light_right",
    "right",
    "hard_right",
    "u_turn_right",
    "u_turn_left",
    "hard_left",
    "left",
    "light_left",
    "bear_left",
]

INSTRUCTION_DIRECTION: bidict[
    InstructionDirectionStr, QtLocation.QGeoManeuver.InstructionDirection
] = bidict(
    none=QtLocation.QGeoManeuver.InstructionDirection.NoDirection,
    forward=QtLocation.QGeoManeuver.InstructionDirection.DirectionForward,
    bear_right=QtLocation.QGeoManeuver.InstructionDirection.DirectionBearRight,
    light_right=QtLocation.QGeoManeuver.InstructionDirection.DirectionLightRight,
    right=QtLocation.QGeoManeuver.InstructionDirection.DirectionRight,
    hard_right=QtLocation.QGeoManeuver.InstructionDirection.DirectionHardRight,
    u_turn_right=QtLocation.QGeoManeuver.InstructionDirection.DirectionUTurnRight,
    u_turn_left=QtLocation.QGeoManeuver.InstructionDirection.DirectionUTurnLeft,
    hard_left=QtLocation.QGeoManeuver.InstructionDirection.DirectionHardLeft,
    left=QtLocation.QGeoManeuver.InstructionDirection.DirectionLeft,
    light_left=QtLocation.QGeoManeuver.InstructionDirection.DirectionLightLeft,
    bear_left=QtLocation.QGeoManeuver.InstructionDirection.DirectionBearLeft,
)


class GeoManeuver(QtLocation.QGeoManeuver):
    """Represents info relevant to the point at which two QGeoRouteSegments meet."""

    def __bool__(self):
        return self.isValid()

    def __str__(self):
        return self.instructionText()

    def __setitem__(self, index: str, val):
        attrs = self.extendedAttributes()
        attrs[index] = val
        self.setExtendedAttributes(attrs)

    def __getitem__(self, index: str):
        attr = self.extendedAttributes()
        if index not in attr:
            raise KeyError(f"Key {index!r} does not exist.")
        return attr[index]

    def get_position(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.position())

    def get_waypoint(self) -> positioning.GeoCoordinate | None:
        wp = self.waypoint()
        return positioning.GeoCoordinate(wp) if wp.isValid() else None

    def set_direction(
        self,
        direction: InstructionDirectionStr | QtLocation.QGeoManeuver.InstructionDirection,
    ):
        """Set the direction.

        Args:
            direction: Direction
        """
        self.setDirection(INSTRUCTION_DIRECTION.get_enum_value(direction))

    def get_direction(self) -> InstructionDirectionStr:
        """Return current direction.

        Returns:
            Direction
        """
        return INSTRUCTION_DIRECTION.inverse[self.direction()]


if __name__ == "__main__":
    maneuver = GeoManeuver()
    maneuver.setExtendedAttributes(dict(a="a"))
    maneuver["test"] = "test"
    waypoint = maneuver.get_waypoint()



from __future__ import annotations

import dataclasses
import sys

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import datatypes


@dataclasses.dataclass
class ActionGridItem:
    button: widgets.ToolButton
    action: gui.QAction
    row: int
    column: int


PM = widgets.QStyle.PixelMetric
CT = widgets.QStyle.ContentsType
CC = widgets.QStyle.ComplexControl


class AcionGridButton(widgets.ToolButton):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._elided_text = ""
        self.set_size_policy("preferred", "preferred")

    def actionEvent(self, event):
        super().actionEvent(event)
        match event.type():
            case core.QEvent.Type.ActionChanged | core.QEvent.Type.ActionAdded:
                self._update_text()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._update_text()

    def add(self, action):
        self.addAction(action)

    def _update_text(self):
        fm = self.get_font_metrics()
        text = self.defaultAction().text()
        words = text.split()

        option = widgets.QStyleOptionToolButton()
        option.initFrom(self)

        margin = self.style().pixelMetric(PM.PM_ButtonMargin, option, self)
        min_width = self.width() - 2 * margin

        lines = []

        if fm.boundingRect(" ".join(words)).width() <= min_width or len(words) <= 1:
            lines = [" ".join(words)]
        else:
            best_w, best_l = sys.maxsize, ["", ""]
            for i in range(1, len(words)):
                l1 = " ".join(words[:i])
                l2 = " ".join(words[i:])
                width = max(fm.boundingRect(l1).width(), fm.boundingRect(l2).width())
                if width < best_w:
                    best_w = width
                    best_l = [l1, l2]
            lines = best_l

        # elide the end of each line if too long
        lines = [fm.elided_text(line, "right", self.width() - margin) for line in lines]
        self._elided_text = "\n".join(lines).replace("&", "&&")  # Need escaped ampersand

    def initStyleOption(self, option):
        super().initStyleOption(option)
        if self._elided_text:
            option.text = self._elided_text

    def sizeHint(self):
        opt = widgets.QStyleOptionToolButton()
        self.initStyleOption(opt)
        style = self.style()
        csize = opt.iconSize
        margin = style.pixelMetric(PM.PM_ButtonMargin)
        # content size is:
        #   * vertical: icon + margin + 2 * font ascent
        #   * horizontal: icon * 3 / 2

        csize.setHeight(csize.height() + margin + 2 * opt.fontMetrics.lineSpacing())
        csize.setWidth(csize.width() * 3 // 2)
        return style.sizeFromContents(CT.CT_ToolButton, opt, csize, self)


class ActionGrid(widgets.Frame):
    """A widget containing a grid of actions/buttons, with a toolbar-like interface.

    Parameters:
        columns : Number of columns in the grid layout.
        button_size : Size of tool buttons in the grid.
        îcon_size : Size of icons in the buttons.
        tool_button_style : Tool button style.
    """

    actionTriggered = core.Signal(gui.QAction)
    actionHovered = core.Signal(gui.QAction)

    def __init__(
        self,
        columns: int = 4,
        button_size: datatypes.SizeType | None = None,
        icon_size: datatypes.SizeType | None = None,
        tool_button_style=constants.ToolButtonStyle.ToolButtonTextUnderIcon,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._columns = columns
        self._button_size = (
            datatypes.to_size(button_size) if button_size else core.QSize()
        )
        self._icon_size = datatypes.to_size(icon_size) if icon_size else core.QSize()
        self._tool_button_style = tool_button_style

        self._grid_slots: list[ActionGridItem] = []
        self._mapper = core.SignalMapper()
        self._mapper.mappedObject.connect(self._on_clicked)

        self.set_layout("grid", margin=0, spacing=0)
        self.set_size_policy("fixed", "minimum_expanding")

    def set_button_size(self, size: core.QSize):
        """Set the button size."""
        if self._button_size != size:
            self._button_size = core.QSize(size)
            for slot in self._grid_slots:
                slot.button.setFixedSize(size)

    def button_size(self) -> core.QSize:
        """Return the button size."""
        return core.QSize(self._button_size)

    def setIconSize(self, size: core.QSize):
        """Set the button icon size (The default icon size is style-defined)."""
        if self._icon_size != size:
            self._icon_size = core.QSize(size)
            size = self._get_effective_icon_size()
            for slot in self._grid_slots:
                slot.button.setIconSize(size)

    def icon_size(self) -> core.QSize:
        """Return icon size. If no size is set a style defined size is returned."""
        return self._get_effective_icon_size()

    def _get_effective_icon_size(self) -> core.QSize:
        if self._icon_size.isValid():
            return core.QSize(self._icon_size)
        opt = widgets.QStyleOptionToolButton()
        opt.initFrom(self)
        s = self.style().pixelMetric(PM.PM_LargeIconSize, opt, None)
        return core.QSize(s, s)

    def changeEvent(self, event):
        if event.type() == core.QEvent.Type.StyleChange:
            size = self._get_effective_icon_size()
            for item in self._grid_slots:
                item.button.setIconSize(size)
        super().changeEvent(event)

    def set_tool_button_style(self, style: constants.ToolButtonStyle):
        """Set the tool button style."""
        if self._tool_button_style != style:
            self._tool_button_style = style
            for slot in self._grid_slots:
                slot.button.setToolButtonStyle(style)

    def get_tool_button_style(self) -> constants.ToolButtonStyle:
        """Return the tool button style."""
        return self._tool_button_style

    def set_column_count(self, columns: int):
        """Set the number of button/action columns."""
        if self._columns != columns:
            self._columns = columns
            self._relayout()

    def columns(self) -> int:
        """Return the number of columns in the grid."""
        return self._columns

    def clear(self):
        """Clear all actions/buttons."""
        for slot in reversed(list(self._grid_slots)):
            self.removeAction(slot.action)
        self._grid_slots = []

    def get_button_for_action(self, action):
        """Return the :class:`QToolButton` instance button for `action`."""
        actions = [slot.action for slot in self._grid_slots]
        index = actions.index(action)
        return self._grid_slots[index].button

    def count(self) -> int:
        """Return the number of buttons/actions in the grid."""
        return len(self._grid_slots)

    def actionEvent(self, event):
        super().actionEvent(event)
        match event.type():
            case core.QEvent.Type.ActionAdded:
                # Note: the action is already in the self.actions() list.
                actions = list(self.actions())
                index = actions.index(event.action())
                self._insert_action_button(index, event.action())

            case core.QEvent.Type.ActionRemoved:
                self._remove_action_button(event.action())

    def _insert_action_button(self, index, action):
        """Create a button for the action and add it to the layout at index."""
        self._shift_grid(index, 1)
        button = AcionGridButton(
            self,
            tool_button_style=self._tool_button_style,
            triggered=self._mapper.map,
            icon_size=self._get_effective_icon_size(),
        )
        button.setDefaultAction(action)

        if self._button_size.isValid():
            button.setFixedSize(self._button_size)
        row = index // self._columns
        column = index % self._columns
        self.layout().addWidget(button, row, column)
        self._grid_slots.insert(index, ActionGridItem(button, action, row, column))
        self._mapper.setMapping(button, action)
        button.installEventFilter(self)

    def _remove_action_button(self, action):
        """Remove the button for the action from the layout and delete it."""
        actions = [slot.action for slot in self._grid_slots]
        index = actions.index(action)
        slot = self._grid_slots.pop(index)

        slot.button.removeEventFilter(self)
        self._mapper.removeMappings(slot.button)

        self.layout().removeWidget(slot.button)
        self._shift_grid(index + 1, -1)

        slot.button.deleteLater()

    def _shift_grid(self, start: int, count: int = 1):
        """Shift all buttons starting at index `start` by `count` cells."""
        layout = self.layout()
        button_count = layout.count()
        columns = self._columns

        direction = 1 if count >= 0 else -1
        if direction == 1:
            start, end = button_count - 1, start - 1
        else:
            start, end = start, button_count

        for index in range(start, end, -direction):
            if item := layout.itemAtPosition(index // columns, index % columns):
                button = item.widget()
                new_index = index + count
                layout.addWidget(button, new_index // columns, new_index % columns)

    def _relayout(self):
        """Relayout the buttons."""
        layout = self.layout()
        layout.clear()
        cols = self._columns
        self._grid_slots = [
            ActionGridItem(slot.button, slot.action, i // cols, i % cols)
            for i, slot in enumerate(self._grid_slots)
        ]
        for slot in self._grid_slots:
            layout.addWidget(slot.button, slot.row, slot.column)

    def _index_of(self, button) -> int:
        """Return the index of button widget."""
        buttons = [slot.button for slot in self._grid_slots]
        return buttons.index(button)

    def _on_button_enter(self, button):
        action = button.defaultAction()
        self.actionHovered.emit(action)

    @core.Slot(core.QObject)
    def _on_clicked(self, action):
        self.actionTriggered.emit(action)

    def eventFilter(self, obj, event) -> bool:
        match event.type():
            case core.QEvent.Type.KeyPress if obj.hasFocus():
                combo = core.KeyCombination(event.keyCombination())
                if combo.is_moving() and self._focus_move(obj, combo.key()):
                    event.accept()
                    return True
            case core.QEvent.Type.HoverEnter if obj.parent() is self:
                self._on_button_enter(obj)
        return super().eventFilter(obj, event)

    def _focus_move(self, focus: widgets.QWidget, key: constants.Key) -> bool:
        assert focus is self.focusWidget()
        try:
            index = self._index_of(focus)
        except IndexError:
            return False
        match key:
            case constants.Key.Key_Down:
                index += self._columns
            case constants.Key.Key_Up:
                index -= self._columns
            case constants.Key.Key_Left:
                index -= 1
            case constants.Key.Key_Right:
                index += 1

        if 0 <= index < self.count():
            button = self._grid_slots[index].button
            button.setFocus(constants.FocusReason.TabFocusReason)
            return True
        else:
            return False


if __name__ == "__main__":
    from prettyqt import gui

    app = widgets.app()
    toolbox = ActionGrid(columns=3)
    w = widgets.Widget()
    w.set_layout("horizontal")
    w.box.add(toolbox)
    icon = app.style().standardIcon(widgets.QStyle.StandardPixmap.SP_FileIcon)
    actions = [
        gui.QAction(text="A", icon=icon, parent=toolbox),
        gui.QAction(text="B", icon=icon, parent=toolbox),
        gui.QAction(text="This one is longer.", icon=icon, parent=toolbox),
        gui.QAction(text="Not done yet!", icon=icon, parent=toolbox),
        gui.QAction(
            text="The quick brown fox ... does something I guess",
            icon=icon,
            parent=toolbox,
        ),
    ]

    toolbox.addActions(actions)
    w.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, core


class TableToListProxyModel(core.IdentityProxyModel):
    """Proxy model to flatten a table to a list.

    Reshapes a table by concatenating all columns into one large column,
    so that the new rowCount equals to sourceModel rowCount * sourceModel columnCount.
    If a verticalHeader is available, it will show the original position of the cell.

    === "Without proxy"

        ```py
        data = dict(
            first=["John", "Mary"],
            last=["Doe", "Bo"],
            height=[5.5, 6.0],
            weight=[130, 150],
        )
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        # table.proxifier.to_list()
        table.show()

        ```
        <figure markdown>
          ![Image title](../../images/tabletolistproxymodel_before.png)
        </figure>

    === "With proxy"

        ```py
        data = dict(
            first=["John", "Mary"],
            last=["Doe", "Bo"],
            height=[5.5, 6.0],
            weight=[130, 150],
        )
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.proxifier.to_list()
        table.show()
        ```
        <figure markdown>
          ![Image title](../../images/tabletolistproxymodel_after.png)
        </figure>

    """

    ID = "table_to_list"
    ICON = "mdi6.table-pivot"

    def __init__(self, *args, header_title: str = "", **kwargs):
        super().__init__(*args, **kwargs)
        self._header_title = header_title

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return 0 if self.sourceModel() is None else 1

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self._header_title or None
            case constants.VERTICAL, constants.DISPLAY_ROLE:
                col_section = section % super().columnCount()
                row_section = section // super().columnCount()
                pre = super().headerData(col_section, constants.HORIZONTAL, role)
                post = super().headerData(row_section, constants.VERTICAL, role)
                pre_str = col_section if pre is None else pre
                post_str = row_section if post is None else post
                return f"{pre_str} | {post_str}"
        return None

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        source = self.sourceModel()
        return 0 if source is None else source.rowCount() * source.columnCount()

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        source = self.sourceModel()
        if row < 0 or column < 0 or source is None:
            return core.ModelIndex()
        source_parent = self.mapToSource(parent)
        colcount = source.columnCount()
        source_index = source.index(row // colcount, row % colcount, source_parent)
        return self.mapFromSource(source_index)

    def mapToSource(self, proxy_idx: core.ModelIndex) -> core.ModelIndex:
        source = self.sourceModel()
        if source is None or not proxy_idx.isValid():
            return core.ModelIndex()
        row = proxy_idx.row()
        colcount = source.columnCount()
        return source.index(row // colcount, row % colcount)

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        source = self.sourceModel()
        if source is None or not source_index.isValid():
            return core.ModelIndex()
        r = source_index.row() * source.columnCount() + source_index.column()
        return self.createIndex(r, 0, source_index.internalPointer())

    def set_header_title(self, title: str):
        self._header_title = title
        self.headerDataChanged.emit(constants.HORIZONTAL, 0, 0)

    def get_header_title(self) -> str:
        return self._header_title

    header_title = core.Property(
        str,
        get_header_title,
        set_header_title,
        doc="Column header for resulting column",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app()
    data = dict(
        first=["John", "Mary"],
        last=["Doe", "Bo"],
        height=[5.5, 6.0],
        weight=[130, 150],
    )
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    table.show()
    table.resize(600, 500)
    table.h_header.resize_sections("stretch")
    table.set_title("Table to list")
    table.set_icon("mdi6.table-pivot")
    # table.proxifier.transpose()
    # table.proxifier.to_list()
    # splitter = debugging.ProxyComparerWidget(table.model())
    # splitter.show()
    app.exec()



from __future__ import annotations

from typing_extensions import Self

from prettyqt.qt import QtCore
from prettyqt.utils import datatypes, get_repr


class SizeF(QtCore.QSizeF):
    """Defines the size of a two-dimensional object using floating point precision."""

    def __repr__(self):
        return get_repr(self, self.width(), self.height())

    @property
    def _width(self) -> float:
        return self.width()

    @property
    def _height(self) -> float:
        return self.height()

    __match_args__ = ("_width", "_height")

    def __getitem__(self, index) -> float:
        return (self.width(), self.height())[index]

    def __reduce__(self):
        return type(self), (self.width(), self.height())

    def expanded_to(self, size: datatypes.SizeFType) -> Self:
        size = datatypes.to_sizef(size)
        return type(self)(self.expandedTo(size))

    def shrunk_by(self, margins: datatypes.MarginsFType) -> Self:
        margins = datatypes.to_marginsf(margins)
        return type(self)(self.shrunkBy(margins))

    def grown_by(self, margins: datatypes.MarginsFType) -> Self:
        margins = datatypes.to_marginsf(margins)
        return type(self)(self.grownBy(margins))


if __name__ == "__main__":
    size = SizeF(10, 20)



from __future__ import annotations

from typing import Literal

from prettyqt import charts, constants, core, gui, widgets
from prettyqt.utils import bidict, datatypes


ThemeStr = Literal[
    "Light",
    "Blue Cerulean",
    "Dark",
    "Brown Sand",
    "Blue NCS",
    "High Contrast",
    "Blue Icy",
    "Qt",
]

THEMES: bidict[ThemeStr, charts.QChart.ChartTheme] = bidict(
    {
        "Light": charts.QChart.ChartTheme.ChartThemeLight,
        "Blue Cerulean": charts.QChart.ChartTheme.ChartThemeBlueCerulean,
        "Dark": charts.QChart.ChartTheme.ChartThemeDark,
        "Brown Sand": charts.QChart.ChartTheme.ChartThemeBrownSand,
        "Blue NCS": charts.QChart.ChartTheme.ChartThemeBlueNcs,
        "High Contrast": charts.QChart.ChartTheme.ChartThemeHighContrast,
        "Blue Icy": charts.QChart.ChartTheme.ChartThemeBlueIcy,
        "Qt": charts.QChart.ChartTheme.ChartThemeQt,
    }
)


AnimationOptionStr = Literal["none", "grid_axis", "series", "all"]

ANIMATION_OPTIONS: bidict[AnimationOptionStr, charts.QChart.AnimationOption] = bidict(
    none=charts.QChart.AnimationOption.NoAnimation,
    grid_axis=charts.QChart.AnimationOption.GridAxisAnimations,
    series=charts.QChart.AnimationOption.SeriesAnimations,
    all=charts.QChart.AnimationOption.AllAnimations,
)

ChartTypeStr = Literal["undefined", "cartesian", "polar"]

CHART_TYPES: bidict[ChartTypeStr, charts.QChart.ChartType] = bidict(
    undefined=charts.QChart.ChartType.ChartTypeUndefined,
    cartesian=charts.QChart.ChartType.ChartTypeCartesian,
    polar=charts.QChart.ChartType.ChartTypePolar,
)


class ChartMixin(widgets.GraphicsWidgetMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_x = 0
        self.max_y = 0
        self.min_x = 0
        self.min_y = 0
        self.adjust_style_to_palette()
        gui.GuiApplication.styleHints().colorSchemeChanged.connect(
            self.adjust_style_to_palette
        )

    def adjust_style_to_palette(self):
        """Adjusts the chart theme to current Palette.

        Checks if palette is dark-ish and applies an appropriate theme to the chart.
        """
        pal = gui.GuiApplication.get_palette()
        style = "Dark" if pal.is_dark() else "Light"
        self.set_theme(style)

    def get_axes(
        self,
        orientation: constants.OrientationStr | constants.Orientation | None = None,
        series: charts.QAbstractBarSeries | None = None,
    ) -> list[charts.QAbstractAxis]:
        """Get axes of chart.

        Arguments:
            orientation: Orientation of the axes that should get returned.
            series: Series to return axes for. Returns all axes if None.
        """
        if orientation is None:
            orientation = constants.HORIZONTAL | constants.VERTICAL
        return self.axes(constants.ORIENTATION.get_enum_value(orientation), series)

    def update_boundaries(self):
        """Set new min/max values based on axis."""
        if axis_x := self.get_axes("horizontal"):
            self.max_x = axis_x[0].max()
            self.min_x = axis_x[0].min()
        if axis_y := self.get_axes("vertical"):
            self.max_y = axis_y[0].max()
            self.min_y = axis_y[0].min()

    def hide_legend(self):
        self.legend().hide()

    def show_legend(self):
        self.legend().show()

    def get_legend(self) -> charts.Legend:
        return charts.Legend(self.legend())

    def set_legend_alignment(
        self, alignment: constants.SideStr | constants.AlignmentFlag
    ):
        """Set alignment of the chart legend."""
        self.legend().setAlignment(constants.SIDES.get_enum_value(alignment))

    def set_theme(self, theme_name: ThemeStr | charts.QChart.ChartTheme):
        self.setTheme(THEMES.get_enum_value(theme_name))

    def set_margins(self, margins: datatypes.MarginsType):
        margins = datatypes.to_margins(margins)
        self.setMargins(margins)

    def set_animation_options(
        self, option: AnimationOptionStr | charts.QChart.AnimationOption
    ):
        self.setAnimationOptions(ANIMATION_OPTIONS.get_enum_value(option))

    def apply_nice_numbers(self):
        """Adjust both axis to display nice round numbers."""
        for axis in self.get_axes():
            axis.applyNiceNumbers()

    def zoom_by_factor(self, factor: float):
        """Zoom in/out by factor (1.0 = no change).

        Make sure that we dont zoom out too far
        """
        self.zoom(factor)
        if axis_x := self.get_axes("horizontal"):
            if axis_x[0].min() < self.min_x:
                axis_x[0].setMin(self.min_x)
            if axis_x[0].max() > self.max_x:
                axis_x[0].setMax(self.max_x)
        if axis_y := self.get_axes("vertical"):
            if axis_y[0].max() > self.max_y:
                axis_y[0].setMax(self.max_y)

            # always bottom-align when zooming for now. should perhaps become optional.
            # if axis_y[0].min() < self.min_y:
            axis_y[0].setMin(max(0, self.min_y))

    def get_chart_type(self) -> ChartTypeStr:
        return CHART_TYPES.inverse[self.chartType()]

    def get_margins(self) -> core.Margins:
        return core.Margins(self.margins())

    def get_plot_area(self) -> core.RectF:
        return core.RectF(self.plotArea())

    def get_locale(self) -> core.Locale:
        return core.Locale(self.locale())

    def get_theme(self) -> ThemeStr:
        return THEMES.inverse[self.theme()]

    def get_animation_options(self) -> list[AnimationOptionStr]:
        return ANIMATION_OPTIONS.get_list(self.animationOptions())

    def get_animation_easing_curve(self) -> core.EasingCurve:
        return core.EasingCurve(self.animationEasingCurve())


class Chart(ChartMixin, charts.QChart):
    pass


if __name__ == "__main__":
    app = widgets.app()
    chart = Chart()
    legend = chart.get_legend()
    legend.set_alignment("bottom")



from __future__ import annotations

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import colors, datatypes


CC_ScrollBar = widgets.QStyle.ComplexControl.CC_ScrollBar
SubControl = widgets.QStyle.SubControl


class AnnotatedScrollBar(widgets.ScrollBar):
    r"""ScrollBar which can highlight user-defined ranges.

    ``` py
    widget = widgets.PlainTextEdit("\n".join(str(i) for i in range(1000)))
    widget.v_scrollbar = AnnotatedScrollBar(constants.VERTICAL)
    widget.v_scrollbar.set_annotations([(10, 20), (50, 60), (82, 85)])
    ```

    <figure markdown>
      ![Image title](../../images/annotatedscrollbar.png)
      <figcaption>Annotated ScrollBar</figcaption>
    </figure>
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._annotation_color = gui.QColor("gold")
        self._annotations = []
        self._document_length = 100

    @classmethod
    def create_example(cls):
        widget = widgets.PlainTextEdit("\n".join(str(i) for i in range(1000)))
        widget.v_scrollbar = AnnotatedScrollBar(constants.VERTICAL)
        widget.v_scrollbar.set_annotations([(10, 20), (50, 60), (82, 85)])
        return widget

    def set_annotations(self, annotations: list[tuple[int, int]]):
        self._annotations = annotations

    def get_annotations(self) -> list[tuple[int, int]]:
        return self._annotations

    def add_annotation(self, annotation: tuple[int, int]):
        self._annotations.append(annotation)

    def paintEvent(self, event):
        super().paintEvent(event)
        with gui.Painter(self) as p:
            opt = widgets.QStyleOptionSlider()
            self.initStyleOption(opt)
            gr = self.style().subControlRect(
                CC_ScrollBar, opt, SubControl.SC_ScrollBarGroove, self
            )
            sr = self.style().subControlRect(
                CC_ScrollBar, opt, SubControl.SC_ScrollBarSlider, self
            )
            p.setClipRegion(
                gui.QRegion(gr) - gui.QRegion(sr), constants.ClipOperation.IntersectClip
            )
            x, y, w, h = gr.getRect()
            c = gui.QColor(self._annotation_color)
            p.setBrush(c)
            c.setAlphaF(0.3)
            p.setPen(gui.QPen(c, 2.0))
            scale = 1.0 / self._document_length
            if self.orientation() == constants.VERTICAL:
                rects = [
                    core.QRect(
                        x + 1,
                        y + h * start * scale - 0.5,
                        w - 2,
                        h * (end - start) * scale + 1,
                    )
                    for start, end in self._annotations
                ]
            else:
                rects = [
                    core.QRect(
                        x + w * start * scale - 0.5,
                        y + 1,
                        w * (end - start) * scale + 1,
                        h - 2,
                    )
                    for start, end in self._annotations
                ]
            p.drawRects(rects)

    def set_annotation_color(self, color: datatypes.ColorType):
        self._annotation_color = colors.get_color(color).as_qt()

    def get_annotation_color(self) -> gui.QColor:
        return self._annotation_color

    def set_document_length(self, length: int):
        self._document_length = length

    def get_document_length(self) -> int:
        return self._document_length

    annotation_color = core.Property(
        gui.QColor,
        get_annotation_color,
        set_annotation_color,
        doc="Color for the annotated regions",
    )
    annotations = core.Property(
        list,
        get_annotations,
        set_annotations,
        doc="Current set of annotations",
    )
    document_length = core.Property(
        int,
        get_document_length,
        set_document_length,
        doc="Total document length",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.PlainTextEdit("\n".join(str(i) for i in range(1000)))
    widget.v_scrollbar = AnnotatedScrollBar(constants.VERTICAL)
    widget.v_scrollbar.set_annotations([(10, 20), (50, 60), (82, 85)])
    widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import datetime

from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict, datatypes


SectionStr = Literal[
    "none",
    "am_pm",
    "msec",
    "second",
    "minute",
    "hour",
    "day",
    "month",
    "year",
]

SECTIONS: bidict[SectionStr, widgets.QDateTimeEdit.Section] = bidict(
    none=widgets.QDateTimeEdit.Section.NoSection,
    am_pm=widgets.QDateTimeEdit.Section.AmPmSection,
    msec=widgets.QDateTimeEdit.Section.MSecSection,
    second=widgets.QDateTimeEdit.Section.SecondSection,
    minute=widgets.QDateTimeEdit.Section.MinuteSection,
    hour=widgets.QDateTimeEdit.Section.HourSection,
    day=widgets.QDateTimeEdit.Section.DaySection,
    month=widgets.QDateTimeEdit.Section.MonthSection,
    year=widgets.QDateTimeEdit.Section.YearSection,
)


class DateTimeEditMixin(widgets.AbstractSpinBoxMixin):
    """Widget for editing dates and times."""

    value_changed = core.Signal(datetime.datetime)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setCalendarPopup(True)
        self.dateTimeChanged.connect(self.datetime_changed)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "timeSpec": constants.TIME_SPEC,
            "currentSection": SECTIONS,
        }
        return maps

    def datetime_changed(self, date):
        dt = self.get_datetime()
        self.value_changed.emit(dt)

    def set_to_today(self):
        self.setDateTime(core.DateTime.currentDateTime())

    def get_section_text(
        self, section: SectionStr | widgets.QDateTimeEdit.Section
    ) -> str:
        return self.sectionText(SECTIONS.get_enum_value(section))

    def get_current_section(self) -> SectionStr:
        return SECTIONS.inverse[self.currentSection()]

    def set_current_section(self, section: SectionStr | widgets.QDateTimeEdit.Section):
        self.setCurrentSection(SECTIONS.get_enum_value(section))

    def get_displayed_sections(self) -> list[SectionStr]:
        return SECTIONS.get_list(self.displayedSections())

    def set_range(
        self,
        lower: datatypes.DateTimeType,
        upper: datatypes.DateTimeType,
    ):
        lower = datatypes.to_datetime(lower)
        upper = datatypes.to_datetime(upper)
        self.setToolTip(f"{lower} <= x <= {upper}")
        self.setDateTimeRange(lower, upper)

    def set_format(self, fmt: str):
        self.setDisplayFormat(fmt)

    def get_value(self) -> datetime.datetime:
        return self.get_datetime()

    def set_value(self, value: datatypes.DateTimeType):
        self.setDateTime(datatypes.to_datetime(value))

    def get_datetime(self) -> datetime.datetime:
        return self.dateTime().toPython()  # type: ignore

    def min_datetime(self) -> datetime.datetime:
        return self.minimumDateTime().toPython()  # type: ignore

    def max_datetime(self) -> datetime.datetime:
        return self.maximumDateTime().toPython()  # type: ignore

    def min_date(self) -> datetime.date:
        return self.minimumDate().toPython()  # type: ignore

    def max_date(self) -> datetime.date:
        return self.maximumDate().toPython()  # type: ignore

    def get_date(self) -> datetime.date:
        return self.date().toPython()  # type: ignore

    def min_time(self) -> datetime.time:
        return self.minimumTime().toPython()  # type: ignore

    def max_time(self) -> datetime.time:
        return self.maximumTime().toPython()  # type: ignore

    def get_time(self) -> datetime.time:
        return self.time().toPython()  # type: ignore


class DateTimeEdit(DateTimeEditMixin, widgets.QDateTimeEdit):
    @classmethod
    def setup_example(cls):
        widget = DateTimeEdit()
        widget.set_value("01.03.1998 02:11")
        return widget


if __name__ == "__main__":
    app = widgets.app()
    widget = DateTimeEdit()
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Iterator

from prettyqt import constants, core, gui, iconprovider
from prettyqt.utils import datatypes, helpers, listdelegators


class StandardItemModel(core.AbstractItemModelMixin, gui.QStandardItemModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setItemPrototype(gui.StandardItem())

    def __getitem__(
        self, index: int | slice | tuple[int | slice, int | slice] | core.QModelIndex
    ) -> gui.QStandardItem | listdelegators.ListDelegator[gui.QStandardItem]:
        rowcount = self.rowCount()
        colcount = self.columnCount()
        match index:
            case int() as row:
                if row < 0:
                    row = row + rowcount
                if index >= self.rowCount():
                    raise IndexError(index)
                return self.item(index)
            case slice():
                return self.__getitem__(index, 0)
            case core.QModelIndex():
                return self.itemFromIndex(index)
            case int() as row, int() as col:
                if row < 0:
                    row = row + rowcount
                if col < 0:
                    col = col + colcount
                return self.item(row, col)
            case (row, col):
                items = [
                    self.item(i, j)
                    for i, j in helpers.iter_positions(row, col, rowcount, colcount)
                ]
                return listdelegators.ListDelegator(items)
            case _:
                raise TypeError(index)

    def __delitem__(self, index: int | tuple[int, int]):
        match index:
            case int():
                item = self.takeRow(index)
            case (int(), int()):
                item = self.takeItem(*index)
        if item is None:
            raise KeyError(index)
        return item

    def __iter__(self) -> Iterator[gui.QStandardItem]:
        return iter(self.get_children())

    def __getstate__(self):
        return dict(items=[self.item(index) for index in range(self.rowCount())])

    def __setstate__(self, state):
        for item in state["items"]:
            self.appendRow([item])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __add__(self, other: str | gui.QStandardItem) -> StandardItemModel:
        match other:
            case gui.QStandardItem() | str():
                self.add(other)
                return self
            case _:
                raise TypeError("wrong type for addition")

    def get_children(self) -> listdelegators.ListDelegator[gui.QStandardItem]:
        items = [self.item(index) for index in range(self.rowCount())]
        return listdelegators.ListDelegator(items)

    def add(self, *item: str | gui.QStandardItem):
        for i in item:
            new_item = gui.StandardItem(i) if isinstance(i, str) else i
            self.appendRow([new_item])

    def find_items(
        self,
        text: str,
        column: int = 0,
        mode: constants.MatchFlagStr | constants.MatchFlag = "exact",
        recursive: bool = False,
        case_sensitive: bool = False,
    ) -> listdelegators.ListDelegator[gui.QStandardItem]:
        flag = constants.MATCH_FLAGS.get_enum_value(mode)
        if recursive:
            flag |= constants.MatchFlag.MatchRecursive
        if case_sensitive:
            flag |= constants.MatchFlag.MatchCaseSensitive
        items = self.findItems(text, flag, column)  # type: ignore
        return listdelegators.ListDelegator(items)

    def add_item(
        self,
        name: str = "",
        icon: datatypes.IconType = None,
        data: dict | None = None,
        foreground: gui.QBrush | None = None,
        background: gui.QBrush | None = None,
        font: gui.QFont | None = None,
        selectable: bool = True,
        enabled: bool = True,
        editable: bool = False,
        status_tip: str = "",
        tool_tip: str = "",
        whats_this: str | None = None,
        # text_alignment: Optional[str] = None,
        checkstate: constants.StateStr | None = None,
        flags: constants.ItemFlag | None = None,
        size_hint: datatypes.SizeType | None = None,
        is_user_type: bool = False,
    ) -> gui.StandardItem:
        item = gui.StandardItem(name)
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            item.setIcon(icon)
        if data is not None:
            for k, v in data.items():
                item.set_data(v, k)
        if foreground is not None:
            item.setForeground(foreground)
        if background is not None:
            item.setBackground(background)
        if font is not None:
            item.setFont(font)
        if flags is not None:
            item.setFlags(flags)
        if enabled:
            item.setEnabled(enabled)
        if editable:
            item.setEditable(editable)
        if selectable:
            item.setSelectable(selectable)
        if status_tip:
            item.setStatusTip(status_tip)
        if tool_tip:
            item.setToolTip(tool_tip)
        if whats_this:
            item.setWhatsThis(whats_this)
        if size_hint is not None:
            item.set_size_hint(size_hint)
        if checkstate is not None:
            item.set_checkstate(checkstate)
        self.appendRow([item])
        return item

    @classmethod
    def from_dict(cls, dct: dict, **kwargs):
        model = cls(**kwargs)
        model.setHorizontalHeaderLabels(list(dct.keys()))
        for column, v in enumerate(dct.values()):
            for row, label in enumerate(v):
                item = gui.StandardItem(str(label))
                item.setData(label, constants.EDIT_ROLE)
                model.setItem(row, column, item)
        return model


if __name__ == "__main__":
    from prettyqt import widgets

    model = StandardItemModel.from_dict(dict(a=[1, 2, 3], b=["d", "e", "f"]))
    app = widgets.app()
    w = widgets.TableView()
    w.set_model(model)
    w.show()
    app.exec()



from __future__ import annotations

import functools


# def fuzzy_search(search_set, text):
#     for i, word in enumerate(text.split(' ')):
#         token = list(word.lower())

#         if i is not 0:
#           search_set = matches

#         matches = []
#         for string in search_set:
#           token_index, string_index = 0, 0
#           matched_positions = []
#           string = string.lower()

#           while string_index < len(string):
#             if string[string_index] == token[token_index]:
#               matched_positions.append(string_index)
#               token_index += 1

#               if token_index >= len(token):
#                 # matches.append({
#                 #     'match': string,
#                 #     'positions': matched_positions
#                 # })
#                 matches.append(string)

#                 break
#             string_index += 1

#     return matches


def bold(text: str) -> str:
    return f"<b>{text}</b>"


def colored(text: str, color: str) -> str:
    return f"<font color={color!r}>{text}</font>"


@functools.cache
def color_text(input_text: str, text: str, color: str, case_sensitive: bool = False):
    """Color first occurences of input_text chars in text with given color."""

    def converter(x):
        return x if case_sensitive else x.lower()

    input_text = input_text if case_sensitive else input_text.lower()
    output_text = ""
    to_color = ""
    for char in text:
        if input_text and converter(char) == input_text[0]:
            to_color += char
            input_text = input_text[1:]
        else:
            if to_color:
                output_text += bold(colored(to_color, color))
                to_color = ""
            output_text += char
    if to_color:
        output_text += bold(colored(to_color, color))
    return output_text


@functools.cache
def fuzzy_match_simple(pattern: str, instring: str, case_sensitive: bool = False) -> bool:
    """Return True if each character in pattern is found in order in instring.

    Arguments:
        pattern: the pattern to be matched
        instring: the containing string to search against
        case_sensitive: whether to match case-sensitively

    Returns:
        True if there is a match, False otherwise
    """
    p_idx, s_idx, p_len, s_len = 0, 0, len(pattern), len(instring)
    if p_len == 0:
        return True
    if not case_sensitive:
        pattern = pattern.lower()
        instring = instring.lower()
    while (p_idx != p_len) and (s_idx != s_len):
        if pattern[p_idx] == instring[s_idx]:
            p_idx += 1
        s_idx += 1
    return p_len != 0 and s_len != 0 and p_idx == p_len


@functools.cache
def fuzzy_match(
    pattern: str,
    instring: str,
    adj_bonus: int = 5,
    sep_bonus: int = 10,
    camel_bonus: int = 10,
    lead_penalty: int = -3,
    max_lead_penalty: int = -9,
    unmatched_penalty: int = -1,
) -> tuple[bool, int]:
    """Return match boolean and match score.

    Arguments:
        pattern: the pattern to be matched
        instring: the containing string to search against
        adj_bonus: bonus for adjacent matches
        sep_bonus: bonus if match occurs after a separator
        camel_bonus: bonus if match is uppercase
        lead_penalty: penalty applied for each letter before 1st match
        max_lead_penalty: maximum total ``lead_penalty``
        unmatched_penalty: penalty for each unmatched letter

    Returns:
        2-tuple with match truthiness at idx 0 and score at idx 1
    """
    score, p_idx, s_idx, p_len = 0, 0, 0, len(pattern)
    prev_match, prev_lower = False, False
    prev_sep = True  # so that matching first letter gets sep_bonus
    best_letter: str = ""
    best_lower: str | None = None
    best_letter_idx: int | None = None
    best_letter_score = 0
    matched_indices: list[int | None] = []

    for s_idx, s_char in enumerate(instring):
        p_char = pattern[p_idx] if (p_idx != p_len) else None
        p_lower = p_char.lower() if p_char else None
        s_lower, s_upper = s_char.lower(), s_char.upper()

        next_match = p_char and p_lower == s_lower
        rematch = best_letter and best_lower == s_lower

        advanced = next_match and best_letter
        p_repeat = best_letter and p_char and best_lower == p_lower

        if advanced or p_repeat:
            score += best_letter_score
            matched_indices.append(best_letter_idx)
            best_letter, best_lower, best_letter_idx = "", None, None
            best_letter_score = 0

        if next_match or rematch:
            new_score = 0

            # apply penalty for each letter before the first match
            # using max because penalties are negative (so max = smallest)
            if p_idx == 0:
                score += max(s_idx * lead_penalty, max_lead_penalty)

            # apply bonus for consecutive matches
            if prev_match:
                new_score += adj_bonus

            # apply bonus for matches after a separator
            if prev_sep:
                new_score += sep_bonus

            # apply bonus across camelCase boundaries
            if prev_lower and s_char == s_upper and s_lower != s_upper:
                new_score += camel_bonus

            # update pattern index iff the next pattern letter was matched
            if next_match:
                p_idx += 1

            # update best letter match (may be next or rematch)
            if new_score >= best_letter_score:
                # apply penalty for now-skipped letter
                if best_letter:
                    score += unmatched_penalty
                best_letter = s_char
                best_lower = best_letter.lower()
                best_letter_idx = s_idx
                best_letter_score = new_score

            prev_match = True

        else:
            score += unmatched_penalty
            prev_match = False

        prev_lower = s_char == s_lower and s_lower != s_upper
        prev_sep = s_char in "_ "

    if best_letter:
        score += best_letter_score
        matched_indices.append(best_letter_idx)

    return p_idx == p_len, score


# def get_best_matches_alt(
#     search_str: str,
#     collection: list[str],
#     # accessor: Callable[[str], Any] = lambda x: x,
#     sort_results: bool = True,
# ) -> Generator[str, None, None]:
#     """Return list of suggestions for matches of search string inside collection.

#     Arguments:
#         search_str:   A partial string which is typically entered by a user.
#         collection:   A collection of strings to get filtered.
#         accessor:     If the `collection` is not an iterable of strings,
#                       then use the accessor to fetch the string that
#                       will be used for fuzzy matching.
#         sort_results: The suggestions are sorted by considering the
#                       smallest contiguous match, followed by where the
#                       match is found in the full string. If two suggestions
#                       have the same rank, they are then sorted
#                       alpha-numerically. This parameter controls the
#                       *last tie-breaker-alpha-numeric sorting*. The sorting
#                       based on match length and position will be intact.

#     Returns:
#         suggestions (generator): A generator object that produces a list of
#             suggestions narrowed down from `collection` using the `input`.
#     """
#     suggestions: list[tuple[int, int, str, str]] = []
#     pat = ".*?".join(map(re.escape, search_str))
#     pat = f"(?=({pat}))"  # lookahead regex to manage overlapping matches
#     regex = re.compile(pat, re.IGNORECASE)
#     for item in collection:
#         if r := list(regex.finditer(accessor(item))):
#             best = min(r, key=lambda x: len(x.group(1)))  # find shortest match
#             suggestions.append((len(best.group(1)), best.start(), accessor(item), item))

#     if sort_results:
#         return (z[-1] for z in sorted(suggestions))
#     else:
#         return (z[-1] for z in sorted(suggestions, key=lambda x: x[:2]))


def accessor(x: tuple[str, int]) -> int:
    return x[1]


def get_best_matches(search_string: str, candidates: list[str]) -> list[tuple[str, int]]:
    """Return sorted list of all matches."""
    results: list[tuple[str, int]] = [
        (candidate, i[1])
        for candidate in candidates
        if (i := fuzzy_match(search_string, candidate))[0]
    ]
    return sorted(results, key=accessor, reverse=True)


if __name__ == "__main__":
    pat = "aab"
    candidates = ["aaaaab", "aacb", "abc", "abbaab"] * 10000
    print(color_text("ac", "abbbbc", "green"))



from __future__ import annotations

import logging

import mknodes

from mknodes.utils import helpers

from prettyqt import core, gui, itemmodels, prettyqtmarkdown, widgets
from prettyqt.utils import classhelpers


SLICE_PROXY_INFO = """This is a [slice proxy](SliceIdentityProxyModel.md)
and can be selectively applied to a model. Read more about
 [slices](https://docs.python.org/3/library/functions.html#slice)."""

RECURSIVE_MODEL_INFO = "Model can be recursive, so be careful with iterating whole tree."

logger = logging.getLogger(__name__)
app = widgets.app()


class PrettyQtClassPage(mknodes.MkClassPage):
    """A ClassPage specifically for Qt-based classes."""

    def _build(self):
        if qt_parent := classhelpers.get_qt_parent_class(self.klass):
            self.append(f"Qt Base Class: {helpers.link_for_class(qt_parent)}")
            self.append(f"Signature: `{qt_parent.__doc__}`")
        super()._build()
        if issubclass(self.klass, itemmodels.SliceIdentityProxyModel):
            self.add_admonition(SLICE_PROXY_INFO)
        if (
            issubclass(self.klass, core.AbstractItemModelMixin)
            and self.klass.IS_RECURSIVE
        ):
            self.add_admonition(RECURSIVE_MODEL_INFO, "warning")
        if (
            issubclass(self.klass, core.AbstractItemModelMixin)
            and self.klass.DELEGATE_DEFAULT is not None
        ):
            msg = f"Recommended delegate: {self.klass.DELEGATE_DEFAULT!r}"
            self.add_admonition(msg)
        if issubclass(self.klass, core.AbstractItemModelMixin) and hasattr(
            self.klass, "SUPPORTS"
        ):
            self.add_admonition(f"Supported data type: `{self.klass.SUPPORTS}`")
        if issubclass(self.klass, core.QObject):
            table = prettyqtmarkdown.PropertyTable(self.klass, header="⌗ Property table")
            self.append(table)
        if hasattr(self.klass, "ID") and issubclass(self.klass, gui.Validator):
            self.append(f"\n\nValidator ID: **{self.klass.ID}**\n\n")
        if hasattr(self.klass, "ID") and issubclass(
            self.klass, widgets.AbstractItemDelegateMixin
        ):
            self.append(f"\n\nDelegate ID: **{self.klass.ID}**\n\n")
        if (
            "setup_example" in self.klass.__dict__
            and "Abstract" not in self.klass.__name__
            and not self.klass.__name__.endswith("Mixin")
        ):
            if widget := self.klass.setup_example():
                self += prettyqtmarkdown.WidgetScreenShot(widget, header="🖼 Screenshot")


if __name__ == "__main__":
    page = PrettyQtClassPage(klass=core.StringListModel)
    print(page.to_markdown())



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


ErrorStr = Literal[
    "none",
    "custom",
    "not_well_formed",
    "premature_end_of_document",
    "unexpected_element",
]

ERROR: bidict[ErrorStr, QtCore.QXmlStreamReader.Error] = bidict(
    none=QtCore.QXmlStreamReader.Error.NoError,
    custom=QtCore.QXmlStreamReader.Error.CustomError,
    not_well_formed=QtCore.QXmlStreamReader.Error.NotWellFormedError,
    premature_end_of_document=QtCore.QXmlStreamReader.Error.PrematureEndOfDocumentError,
    unexpected_element=QtCore.QXmlStreamReader.Error.UnexpectedElementError,
)

TokenTypeStr = Literal[
    "none",
    "invalid",
    "start_document",
    "end_document",
    "start_element",
    "end_element",
    "characters",
    "comment",
    "dtd",
    "entity_reference",
    "processing_instruction",
]

TOKEN_TYPE: bidict[TokenTypeStr, QtCore.QXmlStreamReader.TokenType] = bidict(
    none=QtCore.QXmlStreamReader.TokenType.NoToken,
    invalid=QtCore.QXmlStreamReader.TokenType.Invalid,
    start_document=QtCore.QXmlStreamReader.TokenType.StartDocument,
    end_document=QtCore.QXmlStreamReader.TokenType.EndDocument,
    start_element=QtCore.QXmlStreamReader.TokenType.StartElement,
    end_element=QtCore.QXmlStreamReader.TokenType.EndElement,
    characters=QtCore.QXmlStreamReader.TokenType.Characters,
    comment=QtCore.QXmlStreamReader.TokenType.Comment,
    dtd=QtCore.QXmlStreamReader.TokenType.DTD,
    entity_reference=QtCore.QXmlStreamReader.TokenType.EntityReference,
    processing_instruction=QtCore.QXmlStreamReader.TokenType.ProcessingInstruction,
)


class XmlStreamReader(QtCore.QXmlStreamReader):
    """Fast parser for reading well-formed XML via a simple streaming API."""

    def __iter__(self):
        return self

    def __next__(self):
        while not self.atEnd():
            self.readNext()
            if self.hasError():
                raise RuntimeError(self.get_error())
            return self
        raise StopIteration

    def get_error(self) -> ErrorStr:
        """Return file error status.

        Returns:
            file error status
        """
        return ERROR.inverse[self.error()]

    def get_token_type(self) -> TokenTypeStr:
        """Get the current token type.

        Returns:
            token type
        """
        return TOKEN_TYPE.inverse[self.tokenType()]

    def read_next(self) -> TokenTypeStr:
        """Read the next token and returns its type.

        Returns:
            token type
        """
        return TOKEN_TYPE.inverse[self.readNext()]


if __name__ == "__main__":
    reader = XmlStreamReader("<a><b>fdd</b></a>")
    for elem in reader:
        print(elem.text())
    print(reader.get_error())



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import bidict, datatypes


RemoveBehaviourStr = Literal["left_tab", "right_tab", "previous_tab"]

REMOVE_BEHAVIOUR: bidict[RemoveBehaviourStr, widgets.QTabBar.SelectionBehavior] = bidict(
    left_tab=widgets.QTabBar.SelectionBehavior.SelectLeftTab,
    right_tab=widgets.QTabBar.SelectionBehavior.SelectRightTab,
    previous_tab=widgets.QTabBar.SelectionBehavior.SelectPreviousTab,
)

ShapeStr = Literal[
    "rounded_north",
    "rounded_south",
    "rounded_west",
    "rounded_east",
    "triangular_north",
    "triangular_south",
    "triangular_west",
    "triangular_east",
]

SHAPE: bidict[ShapeStr, widgets.QTabBar.Shape] = bidict(
    rounded_north=widgets.QTabBar.Shape.RoundedNorth,
    rounded_south=widgets.QTabBar.Shape.RoundedSouth,
    rounded_west=widgets.QTabBar.Shape.RoundedWest,
    rounded_east=widgets.QTabBar.Shape.RoundedEast,
    triangular_north=widgets.QTabBar.Shape.TriangularNorth,
    triangular_south=widgets.QTabBar.Shape.TriangularSouth,
    triangular_west=widgets.QTabBar.Shape.TriangularWest,
    triangular_east=widgets.QTabBar.Shape.TriangularEast,
)

PositionStr = Literal["left", "right"]

POSITIONS: bidict[PositionStr, widgets.QTabBar.ButtonPosition] = bidict(
    left=widgets.QTabBar.ButtonPosition.LeftSide,
    right=widgets.QTabBar.ButtonPosition.RightSide,
)


class TabBarWrapper:
    def __init__(self, indexer, widget: widgets.QTabBar):
        self._widget = widget
        self._range = (
            range(
                indexer.start or 0,
                indexer.stop or self._widget.count(),
                indexer.step or 1,
            )
            if isinstance(indexer, slice)
            else range(indexer, indexer + 1)
        )

    def __getattr__(self, val: str):
        method = self._widget.__getattr__(val)

        if len(self._range) == 1:

            def fn(*args, **kwargs):
                return method(self._range.start, *args, **kwargs)

            return fn

        def fn(*args, **kwargs):
            return [method(i, *args, **kwargs) for i in self._range]

        return fn


class TabBarMixin(widgets.WidgetMixin):
    tab_doubleclicked = core.Signal(int, core.QPoint)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setAcceptDrops(True)
        self.set_elide_mode("right")
        self.set_selection_behavior_on_remove("left_tab")

    def __getitem__(self, index: int | slice):
        return TabBarWrapper(index, self)

    def __setitem__(self, index: tuple[int, PositionStr], value: widgets.QWidget | None):
        self.set_tab(index[0], index[1], value)

    def get_tab_button(
        self, index: int, position: PositionStr | widgets.QTabBar.ButtonPosition
    ) -> widgets.QWidget:
        return self.tabButton(index[0], POSITIONS.get_enum_value(index[1]))

    #  Send the tab_doubleclicked when a tab is double clicked
    def mouseDoubleClickEvent(self, event):
        event.accept()
        tab = self.tabAt(event.position().toPoint())
        pos = gui.QCursor.pos()
        self.tab_doubleclicked.emit(tab, pos)

    def set_icon_size(self, size: datatypes.SizeType):
        """Set size of the icons."""
        self.setIconSize(datatypes.to_size(size))

    def get_icon_size(self) -> core.Size:
        return core.Size(self.iconSize())

    def set_tab(
        self,
        index: int,
        position: PositionStr | widgets.QTabBar.ButtonPosition,
        widget: widgets.QWidget | None,
    ) -> None:
        self.setTabButton(index, POSITIONS.get_enum_value(position), widget)

    def set_selection_behavior_on_remove(
        self, mode: RemoveBehaviourStr | widgets.QTabBar.SelectionBehavior
    ) -> None:
        """Set the remove hehaviour.

        What tab should be set as current when removeTab is called
        if the removed tab is also the current tab.

        Args:
            mode: new remove behaviour
        """
        self.setSelectionBehaviorOnRemove(REMOVE_BEHAVIOUR.get_enum_value(mode))

    def get_remove_behaviour(self) -> RemoveBehaviourStr:
        """Return remove behaviour.

        Returns:
            remove behaviour
        """
        return REMOVE_BEHAVIOUR.inverse[self.selectionBehaviorOnRemove()]

    def set_elide_mode(
        self, mode: constants.TextElideModeStr | constants.TextElideMode
    ) -> None:
        """Set elide mode.

        Args:
            mode: elide mode to use
        """
        self.setElideMode(constants.TEXT_ELIDE_MODE.get_enum_value(mode))

    def get_elide_mode(self) -> constants.TextElideModeStr:
        """Return elide mode.

        Returns:
            elide mode
        """
        return constants.TEXT_ELIDE_MODE.inverse[self.elideMode()]


class TabBar(TabBarMixin, widgets.QTabBar):
    """Tab bar, e.g. for use in tabbed dialogs."""


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    tabwidget = widgets.TabWidget()
    widget2 = widgets.RadioButton("Test")
    widget3 = widgets.PlainTextEdit("Test 243434")
    tabwidget.add_tab(widget2, label="test")
    tabwidget.add_tab(widget3, label="test")
    tabbar = tabwidget.tabBar()
    tabwidget.show()
    app.exec()



from __future__ import annotations

import logging

import mknodes

from prettyqt import constants, core, widgets
from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class WidgetScreenShot(mknodes.MkBinaryImage):
    """Node to easily create and embed a widget screenshot into docs."""

    def __init__(
        self,
        widget: widgets.QWidget,
        caption: str = "",
        path: str | None = None,
        title: str = "Image title",
        header: str = "",
        resize_to: datatypes.SizeType | None = None,
    ):
        if path is None:
            path = f"{widget.__class__.__name__}_screenshot.png"
        logger.info(f"Screenshot for {widget}")
        widget.setAttribute(constants.WidgetAttribute.WA_DontShowOnScreen)
        widget.show()
        widget.adjustSize()
        widgets.app().processEvents()
        pixmap = widget.grab()
        widget.hide()
        widget.deleteLater()
        widgets.app().processEvents()
        ba = core.ByteArray()
        buffer = core.QBuffer(ba)
        buffer.open(core.QIODeviceBase.OpenModeFlag.WriteOnly)
        pixmap.save(buffer, "PNG")
        super().__init__(
            data=ba.data(),
            path=path,
            header=header,
            caption=caption,
            title=title,
        )


if __name__ == "__main__":
    app = widgets.app()

    w = widgets.PushButton("Test")
    page = mknodes.MkPage()
    page += WidgetScreenShot(w, "test.png")
    print(page.to_markdown())



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


MsgTypeStr = Literal["critical", "debug", "info", "warning", "fatal"]

MSG_TYPE: bidict[MsgTypeStr, QtCore.QtMsgType] = bidict(
    critical=QtCore.QtMsgType.QtCriticalMsg,
    debug=QtCore.QtMsgType.QtDebugMsg,
    info=QtCore.QtMsgType.QtInfoMsg,
    warning=QtCore.QtMsgType.QtWarningMsg,
)


class LoggingCategory(QtCore.QLoggingCategory):
    """Represents a category, or 'area' in the logging infrastructure."""

    def get_level(self) -> MsgTypeStr:
        if self.isDebugEnabled():
            return "debug"
        elif self.isInfoEnabled():
            return "info"
        elif self.isWarningEnabled():
            return "warning"
        elif self.isCriticalEnabled():
            return "critical"
        else:
            return "fatal"

    def set_enabled(
        self,
        level: MsgTypeStr,
    ):
        self.setEnabled(MSG_TYPE[level], True)

    def set_disabled(self, level: MsgTypeStr):
        self.setEnabled(MSG_TYPE[level], False)


if __name__ == "__main__":
    cat = LoggingCategory("test")
    cat.set_disabled("debug")
    print(cat.get_level())



from __future__ import annotations

from prettyqt import widgets


class HBoxLayout(widgets.boxlayout.BoxLayoutMixin, widgets.QHBoxLayout):
    """Lines up widgets horizontally."""

    ID = "horizontal"


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    layout = HBoxLayout()
    widget = widgets.Widget()
    widget2 = widgets.RadioButton("Test")
    layout.add(widget2)
    widget.set_layout(layout)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, iconprovider, widgets
from prettyqt.utils import datatypes


class PopupInfo(widgets.Dialog):
    """Dialog overlay to show some info to user."""

    _singleton: PopupInfo | None = None

    def __init__(self, parent: widgets.QWidget | None = None, text: str | None = None):
        super().__init__(parent=parent)
        self.timer = core.Timer(single_shot=True, timeout=self.close)
        self.label = widgets.Label(alignment="center")
        self.iconlabel = widgets.Label()
        self.set_icon("mdi.information")
        self.set_flags(stay_on_top=True, frameless=True, tooltip=True)
        layout = self.set_layout("horizontal", margin=20, size_constraint="minimum")
        self.set_background_color("black")
        self.label.set_color("white")
        layout.add(self.iconlabel)
        layout.add(self.label)
        self.hide()

    def set_text_color(self, color: datatypes.ColorType):
        self.label.set_color(color)

    def set_icon(self, icon: datatypes.IconType):
        pixmap = iconprovider.get_icon(icon, color="white").pixmap(32)
        self.iconlabel.setPixmap(pixmap)

    def show_popup(
        self,
        text: str,
        icon: datatypes.IconType = "mdi.information",
        text_color: datatypes.ColorType = "white",
        background_color: datatypes.ColorType = "black",
        position="screen",
        timeout: int = 2500,
    ):
        self.set_background_color(background_color)
        self.label.set_color(text_color)
        self.label.setText(text)
        self.set_icon(icon)
        self.adjustSize()
        self.position_on(position, y_offset=-200)
        self.timer.start(timeout)
        self.show()

    @classmethod
    def popup(
        cls,
        text: str,
        icon: datatypes.IconType = "mdi.information",
        text_color: datatypes.ColorType = "white",
        background_color: datatypes.ColorType = "black",
        position="screen",
        timeout: int = 2500,
    ):
        if cls._singleton is None:
            cls._singleton = cls()
        popup = cls._singleton
        popup.set_background_color(background_color)
        popup.label.set_color(text_color)
        popup.label.setText(text)
        popup.set_icon(icon)
        popup.adjustSize()
        popup.position_on(position, y_offset=-200)
        popup.timer.start(timeout)
        popup.show()


if __name__ == "__main__":
    app = widgets.app()
    widget = PopupInfo()
    widget.popup("2fdsfsdfsdest")
    app.sleep(1)
    widget.popup("43424234234234324")
    app.exec()



from __future__ import annotations

from typing import Literal

from typing_extensions import Self

from prettyqt import core
from prettyqt.utils import bidict, get_repr


JsonFormatStr = Literal["indented", "compact"]


JSON_FORMAT: bidict[JsonFormatStr, core.QJsonDocument.JsonFormat] = bidict(
    indented=core.QJsonDocument.JsonFormat.Indented,
    compact=core.QJsonDocument.JsonFormat.Compact,
)


class JsonDocument(core.QJsonDocument):
    """Way to read and write JSON documents."""

    def __str__(self):
        return str(self.toVariant())

    def __format__(self, fmt):
        if fmt in JSON_FORMAT:
            return self.to_string(fmt == "indented")
        return super().__format__(fmt)

    def __repr__(self):
        return get_repr(self, self.toVariant())

    def __getitem__(self, index: int | str):
        val = self.array() if self.isArray() else self.object()
        return core.JsonValue(val[index])  # type: ignore

    def __setitem__(self, index: int | str, value):
        if self.isArray():
            if not isinstance(index, int):
                raise TypeError
            array = self.array()
            array[index] = value
            self.setArray(array)
        elif self.isObject():
            if not isinstance(index, str):
                raise TypeError
            obj = self.object()
            obj[index] = value
            self.setObject(obj)

    def to_string(self, indented: bool = False) -> str:
        flag = self.JsonFormat.Indented if indented else self.JsonFormat.Compact
        return self.toJson(flag).data().decode()

    @classmethod
    def from_variant(cls, obj) -> Self:
        doc = cls.fromVariant(obj)
        new = cls()
        if doc.isArray():
            new.setArray(doc.array())
        else:
            new.setObject(doc.object())
        return new


if __name__ == "__main__":
    doc = JsonDocument.from_variant(dict(a="b"))
    new = JsonDocument()
    new.setObject(doc.object())
    new["k"] = "v"



from __future__ import annotations

from collections.abc import Iterator

from prettyqt import gui
from prettyqt.utils import get_repr


class TextBlockGroup(gui.TextObjectMixin, gui.QTextBlockGroup):
    def __repr__(self):
        return get_repr(self)

    def __iter__(self) -> Iterator[gui.TextBlock]:
        return iter(gui.TextBlock(i) for i in self.blockList())

    def get_blocklist(self) -> list[gui.TextBlock]:
        return [gui.TextBlock(i) for i in self.blockList()]


if __name__ == "__main__":
    doc = gui.TextDocument()
    group = TextBlockGroup(doc)



from __future__ import annotations

from prettyqt import constants, core


class ColumnOrderProxyModel(core.IdentityProxyModel):
    """Proxy model which reorders the columns of the source model.

    Proxy model which reorders / hides the columns of the source model by passing a list
    containing the new order. Order indexes can either be an integer or the Column header.
    If not all indexes are part of the list, then the missing sections will be hidden.

    ### Example

    ```py
    table.proxifier.reorder_columns(order=[3, 2, 0])
    table.show()
    # or
    model = MyModel()
    proxy = ColumnOrderProxyModel(order=[3, 2, 0])
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "column_order"
    ICON = "mdi.reorder-vertical"

    def __init__(self, order: list[int | str], **kwargs):
        self._column_order = order
        super().__init__(**kwargs)
        self.set_column_order(order)

    def get_column_order(self) -> list[int]:
        return self._column_order

    def setSourceModel(self, model):
        super().setSourceModel(model)
        self._resolve_string_indexes()

    def _resolve_string_indexes(self):
        new_order = []
        source = self.sourceModel()
        for index in self._column_order:
            if isinstance(index, str):
                for i in range(source.columnCount()):
                    v = source.headerData(i, constants.HORIZONTAL, constants.DISPLAY_ROLE)
                    if v == index:
                        index = i
                        break
                else:
                    raise ValueError(index)
            new_order.append(index)
        self._column_order = new_order

    def set_column_order(self, order: list[int | str]):
        with self.reset_model():
            self._column_order = order
            if self.sourceModel() is not None:
                self._resolve_string_indexes()

    def mapToSource(self, proxy_index: core.ModelIndex) -> core.ModelIndex:
        if not proxy_index.isValid():
            return core.ModelIndex()
        return self.sourceModel().createIndex(
            proxy_index.row(),
            self._column_order[proxy_index.column()],
            proxy_index.internalPointer(),
        )

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        if not source_index.isValid():
            return core.ModelIndex()
        proxy_column = self._column_order.index(source_index.column())
        return self.createIndex(
            source_index.row(), proxy_column, source_index.internalPointer()
        )

    def sibling(self, row: int, column: int, index: core.ModelIndex) -> core.ModelIndex:
        if column >= len(self._column_order):
            return core.ModelIndex()
        return self.index(row, column, index.parent())

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if orientation == constants.HORIZONTAL:
            section = self._column_order[section]
        return self.sourceModel().headerData(section, orientation, role)

    def hasChildren(self, parent_index: core.ModelIndex) -> bool:
        if self.sourceModel() is None or parent_index.column() > 0:
            return False
        source_parent = self.mapToSource(parent_index).sibling(parent_index.row(), 0)
        return self.sourceModel().rowCount(source_parent) > 0

    def parent(self, index: core.ModelIndex) -> core.ModelIndex:
        source_parent = self.mapToSource(index).parent()
        if not source_parent.isValid():
            return core.ModelIndex()
        return self.createIndex(source_parent.row(), 0, source_parent.internalPointer())

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        if parent.column() > 0 or self.sourceModel() is None:
            return core.ModelIndex()
        source_parent = self.mapToSource(parent).sibling(parent.row(), 0)
        source_index = self.sourceModel().index(
            row, self._column_order[column], source_parent
        )
        if not source_index.isValid():
            return core.ModelIndex()
        return self.createIndex(row, column, source_index.internalPointer())

    def columnCount(self, index: core.ModelIndex | None = None) -> int:
        index = index or core.ModelIndex()
        return len(self._column_order)

    def rowCount(self, index: core.ModelIndex | None = None) -> int:
        index = index or core.ModelIndex()
        if self.sourceModel() is None or index.column() > 0:
            return 0
        source_parent = self.mapToSource(index).sibling(index.row(), 0)
        return self.sourceModel().rowCount(source_parent)

    order = core.Property(
        list,
        get_column_order,
        set_column_order,
        doc="Column order",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    data = dict(
        a=["abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedfaa"],
        b=[10000000, 2, 3, 4, 5, 6],
        c=[1, 2, 3, 4, 5, 6],
        d=[100000000, 2, 3, 4, 5, 6],
        e=[1000000, 2, 3, 4, 5, 6],
    )
    source_model = gui.StandardItemModel.from_dict(data)
    app = widgets.app()
    table = widgets.TableView()
    model = ColumnOrderProxyModel(
        order=["b", "c"], parent=table, source_model=source_model
    )
    table.set_model(model)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from collections.abc import Callable
import os
import pathlib

from typing import Literal

from prettyqt import constants, core, gui
from prettyqt.utils import bidict, datatypes, get_repr


AncestorModeStr = Literal["exclude_transients", "include_transients"]

ANCESTER_MODES: bidict[AncestorModeStr, gui.QWindow.AncestorMode] = bidict(
    exclude_transients=gui.QWindow.AncestorMode.ExcludeTransients,
    include_transients=gui.QWindow.AncestorMode.IncludeTransients,
)

VisibilityStr = Literal[
    "windowed", "minimized", "maximized", "fullscreen", "automatic", "hidden"
]

VISIBILITY: bidict[VisibilityStr, gui.QWindow.Visibility] = bidict(
    windowed=gui.QWindow.Visibility.Windowed,
    minimized=gui.QWindow.Visibility.Minimized,
    maximized=gui.QWindow.Visibility.Maximized,
    fullscreen=gui.QWindow.Visibility.FullScreen,
    automatic=gui.QWindow.Visibility.AutomaticVisibility,
    hidden=gui.QWindow.Visibility.Hidden,
)


class WindowMixin(core.ObjectMixin, gui.SurfaceMixin):
    def __repr__(self):
        return get_repr(self)

    def add_shortcut(
        self,
        keysequence: datatypes.KeyCombinationType,
        callback: Callable | None = None,
        context: constants.ShortcutContextStr | constants.ShortcutContext = "window",
    ) -> gui.Shortcut:
        if not isinstance(keysequence, gui.QKeySequence):
            keysequence = gui.KeySequence(keysequence)
        context = constants.SHORTCUT_CONTEXT.get_enum_value(context)
        return gui.Shortcut(keysequence, self, callback, context=context)

    def set_visibility(self, visibility: VisibilityStr | gui.QWindow.Visibility):
        """Set window visibility.

        Args:
            visibility: window visibility
        """
        self.setVisibility(VISIBILITY.get_enum_value(visibility))

    def get_visibility(self) -> VisibilityStr:
        """Get the current window visibility.

        Returns:
            window visibility
        """
        return VISIBILITY.inverse[self.visibility()]

    def start_system_resize(self, edge: constants.EdgeStr | constants.Edge) -> bool:
        """Start system resize.

        Args:
            edge: edge to resize
        """
        return self.startSystemResize(constants.EDGES.get_enum_value(edge))

    def get_screen(self) -> gui.Screen:
        return gui.Screen(self.screen())

    def get_cursor(self) -> gui.Cursor:
        return gui.Cursor(self.cursor())

    def get_icon(self) -> gui.Icon | None:
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def set_modality(
        self, modality: constants.WindowModalityStr | constants.WindowModality
    ) -> None:
        """Set modality for the window.

        Args:
            modality: modality for the window
        """
        self.setModality(constants.WINDOW_MODALITY.get_enum_value(modality))

    def get_modality(self) -> constants.WindowModalityStr:
        return constants.WINDOW_MODALITY.inverse[self.modality()]

    def set_file_path(self, file_path: os.PathLike):
        path = os.fspath(file_path)
        self.setFilePath(path)

    def get_file_path(self) -> pathlib.Path:
        return pathlib.Path(self.filePath())

    def get_type(self) -> constants.WindowTypeStr:
        return constants.WINDOW_TYPE.inverse[self.type()]

    def get_window_state(self) -> constants.WindowStateStr:
        return constants.WINDOW_STATES.inverse[self.windowState()]

    def get_window_states(self) -> list[constants.WindowStateStr]:
        return constants.WINDOW_STATES.get_list(self.windowStates())


class Window(WindowMixin, gui.QWindow):
    pass


if __name__ == "__main__":
    wnd = Window()



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


class VerticalLabel(widgets.Label):
    def __init__(
        self,
        text: str = "",
        orientation: constants.Orientation | constants.OrientationStr = "vertical",
        force_width: bool = True,
        **kwargs,
    ):
        super().__init__(text, **kwargs)
        self.force_width = force_width
        self.orientation = None
        self._size_hint = None
        self.setOrientation(orientation)

    def setOrientation(self, o: constants.Orientation | constants.OrientationStr):
        o = constants.ORIENTATION.get_enum_value(o)
        if self.orientation == o:
            return
        self.orientation = o
        if o == constants.VERTICAL:
            self.set_min_width(0)
            self.set_max_height(None)
        else:
            self.set_min_width(0)
            self.set_max_width(None)
        self.update()
        self.updateGeometry()

    def paintEvent(self, ev):
        with gui.Painter(self) as p:
            if self.orientation == constants.VERTICAL:
                p.rotate(-90)
                rgn = core.QRect(-self.height(), 0, self.height(), self.width())
            else:
                rgn = self.contentsRect()
            align = self.alignment()
            self._size_hint = p.drawText(rgn, align, self.text())

        if self.orientation == constants.VERTICAL:
            self.setMaximumWidth(self._size_hint.height())
            self.setMinimumHeight(self._size_hint.width() if self.force_width else 0)
        else:
            self.setMaximumHeight(self._size_hint.height())
            self.setMinimumWidth(self._size_hint.width() if self.force_width else 0)

    def sizeHint(self):
        if self.orientation == constants.VERTICAL:
            if self._size_hint:
                return core.QSize(self._size_hint.height(), self._size_hint.width())
            else:
                return core.QSize(19, 50)
        else:
            if self._size_hint:
                return core.QSize(self._size_hint.width(), self._size_hint.height())
            else:
                return core.QSize(50, 19)


if __name__ == "__main__":
    app = widgets.app()
    widget = VerticalLabel("tefsdfsdfsdfsdfsdfst", tool_tip="testus")
    widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtTest


class SignalSpy(core.ObjectMixin, QtTest.QSignalSpy):
    pass


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    cb = widgets.CheckBox()
    tester = SignalSpy(cb.clicked)
    cb.animateClick()



from __future__ import annotations

from prettyqt.qt import QtLocation


class PlaceRatings(QtLocation.QPlaceRatings):
    """Holds rating information about a place."""

    def __bool__(self):
        return not self.isEmpty()

    def __float__(self):
        return self.average()


if __name__ == "__main__":
    ratings = PlaceRatings()
    print(bool(ratings))



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping
import contextlib
import logging
import os
import pathlib
import sys
import timeit

import qstylizer.parser
import qstylizer.style

from prettyqt import constants, core, gui, iconprovider, paths, widgets
from prettyqt.utils import datatypes, listdelegators


logger = logging.getLogger(__name__)


SAVE_STATES = dict(
    splitters=widgets.QSplitter,
    mainwindows=widgets.QMainWindow,
    headerviews=widgets.QHeaderView,
)


def setup_runner():
    import inspect

    from prettyqt.qt.QtCore import SignalInstance
    from prettyqt.utils import asyncrunner

    old_connect = SignalInstance.connect
    runner = asyncrunner.AsyncRunner()

    def connect(self, slot):
        if inspect.iscoroutinefunction(slot):
            return old_connect(self, runner.to_sync(slot))
        else:
            return old_connect(self, slot)

    SignalInstance.connect = connect
    return runner


class ApplicationMixin(gui.GuiApplicationMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._debug = False
        self.runner = setup_runner()

    def __iter__(self) -> Iterator[widgets.QWidget]:
        return iter(self.topLevelWidgets())

    def run_parallel(self, fns: Iterable[Callable]):
        self.runner.run_parallel(fns)

    def is_debug(self) -> bool:
        return self._debug

    @contextlib.contextmanager
    def debug_mode(self, log_level: int = logging.DEBUG):
        from prettyqt.debugging import ErrorMessageBox, MessageHandler
        from prettyqt.eventfilters import debugmode

        handler = logging.StreamHandler(sys.stdout)
        f_format = logging.Formatter(
            "%(asctime)s: %(filename)s:%(lineno)d - %(levelname)s - %(message)s"
        )
        handler.setFormatter(f_format)
        handler.setLevel(log_level)
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        root_logger.setLevel(log_level)
        _original_excepthook = sys.excepthook
        sys.excepthook = ErrorMessageBox._excepthook
        eventfilter = debugmode.DebugMode(self)
        self._debug = True
        self.installEventFilter(eventfilter)
        with MessageHandler(root_logger):
            yield self
        self._debug = False
        self.removeEventFilter(eventfilter)
        sys.excepthook = _original_excepthook

    def store_widget_states(
        self, settings: MutableMapping | None = None, key: str = "states"
    ):
        settings = core.Settings() if settings is None else settings
        result = {
            k: {
                i.objectName(): i.saveState()
                for i in self.find_children(v)
                if i.objectName()
            }
            for k, v in SAVE_STATES.items()
        }
        settings[key] = result

    def restore_widget_states(self, settings: Mapping | None = None, key: str = "states"):
        settings = core.Settings() if settings is None else settings
        for category, v in SAVE_STATES.items():
            items = settings[key].get(category)
            if items is None:
                continue
            for name, state in items.items():
                w = self.find_child(v, name=name)
                if w is not None:
                    new_state = state.encode() if isinstance(state, str) else state
                    w.restoreState(new_state)

    @classmethod
    def widget_at(
        cls, pos: datatypes.PointType, typ: type | None = None
    ) -> widgets.QWidget:
        if typ is None:
            return super().widgetAt(pos)
        for widget in cls.widgets_at(pos):
            if isinstance(widget, typ):
                return widget

    @classmethod
    def widgets_at(
        cls, pos: datatypes.PointType
    ) -> listdelegators.ListDelegator[widgets.QWidget]:
        """Return ALL widgets at `pos`.

        Arguments:
            pos: Position at which to get widgets

        Returns:
            list of widgets at given position

        """
        widgets = []
        while widget_at := cls.widgetAt(datatypes.to_point(pos)):
            widgets.append(widget_at)
            # Make widget invisible to further enquiries
            widget_at.setAttribute(constants.WidgetAttribute.WA_TransparentForMouseEvents)
        # Restore attribute
        for widget in widgets:
            widget.setAttribute(
                constants.WidgetAttribute.WA_TransparentForMouseEvents, False
            )

        return widgets

    @classmethod
    def get_mainwindow(cls) -> widgets.QMainWindow | None:
        app = cls.instance()
        if app is None:
            return None
        return next(
            (w for w in app.topLevelWidgets() if isinstance(w, widgets.QMainWindow)),
            None,
        )

    @classmethod
    def get_widget(cls, name: str) -> widgets.QWidget | None:
        mw = cls.get_mainwindow()
        if mw is None:
            logger.warning("Trying to get widget from nonexistent mainwindow")
            return None
        return mw.findChild(widgets.QWidget, name)  # type: ignore
        # widget_list = cls.instance().allWidgets()
        # for widget in widget_list:
        #     if isinstance(widget, widgets.QWidget) and widget.objectName() == name:
        #         return widget
        # return None

    @contextlib.contextmanager
    def edit_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        ss = self.get_stylesheet()
        yield ss
        self.set_stylesheet(ss)

    def set_stylesheet(
        self, ss: None | str | qstylizer.style.StyleSheet | datatypes.PathType
    ):
        match ss:
            case str():
                pass
            case os.PathLike():
                ss = pathlib.Path(ss).read_text()
            case None:
                ss = ""
            case qstylizer.style.StyleSheet():
                ss = str(ss)
            case _:
                raise TypeError(ss)
        self.setStyleSheet(ss)

    def set_style(self, style: str):
        self.setStyle(widgets.QStyleFactory.create(style))
        icon_color = self.get_palette().get_color("highlighted_text")
        iconprovider.set_defaults(color=icon_color)

    def get_stylesheet(self) -> qstylizer.style.StyleSheet:
        return qstylizer.parser.parse(self.styleSheet())

    def set_theme(self, theme: constants.ThemeStr):
        self.set_palette(theme)
        match theme:
            case "default":
                self.set_stylesheet("")
                color = self.get_palette().get_color("highlighted_text")
            case "dark":
                ss = (paths.THEMES_PATH / "darktheme.qss").read_text()
                self.set_stylesheet(ss)
                color = gui.Color("lightblue")
        iconprovider.set_defaults(color=color)

    @classmethod
    def get_available_themes(cls) -> dict[constants.ThemeStr, str]:
        return dict(default="Default", dark="Dark")

    def send_event(self, obj_or_str: str | core.QObject, event: core.QEvent) -> bool:
        obj = self.get_widget(obj_or_str) if isinstance(obj_or_str, str) else obj_or_str
        if obj is None:
            raise ValueError(obj)
        return self.sendEvent(obj, event)

    def post_event(
        self,
        obj_or_str: str | core.QObject,
        event: core.QEvent,
        priority: int | constants.EventPriorityStr = "normal",
    ):
        obj = self.get_widget(obj_or_str) if isinstance(obj_or_str, str) else obj_or_str
        if obj is None:
            raise ValueError(obj)
        super().post_event(obj, event, priority)

    @classmethod
    def get_style_icon(
        cls, icon: widgets.style.StandardPixmapStr | widgets.QStyle.StandardPixmap
    ) -> gui.Icon:
        style = cls.style()
        # icon_size = style.pixelMetric(widgets.QStyle.PM_MessageBoxIconSize)
        icon = style.standardIcon(widgets.style.STANDARD_PIXMAP.get_enum_value(icon))
        return gui.Icon(icon)

    def set_effect_enabled(
        self, effect: constants.UiEffectStr | constants.UIEffect, enabled: bool = True
    ):
        """Set the enabled state of a desktop effect.

        Args:
            effect: desktop effect to set
            enabled: new state
        """
        self.setEffectEnabled(constants.UI_EFFECTS.get_enum_value(effect))

    def is_effect_enabled(self, effect: constants.UiEffectStr) -> bool:
        """Return desktop effect state.

        Returns:
            desktop effect state
        """
        return self.isEffectEnabled(constants.UI_EFFECTS[effect])

    def set_navigation_mode(
        self, mode: constants.NavigationModeStr | constants.NavigationMode
    ):
        """Set the navigation mode.

        Args:
            mode: navigation mode to use
        """
        self.setNavigationMode(constants.NAVIGATION_MODES.get_enum_value(mode))

    def get_navigation_mode(self) -> constants.NavigationModeStr:
        """Return navigation mode.

        Returns:
            navigation mode
        """
        return constants.NAVIGATION_MODES.inverse[self.navigationMode()]

    @classmethod
    def sleep(cls, secs: float):
        """Pause application (non-blocking).

        Args:
            secs: seconds to sleep
        """
        start = timeit.default_timer()
        while timeit.default_timer() - start < secs:
            cls.processEvents()

    @classmethod
    def process_events(
        cls,
        user_input: bool = True,
        socket_notifiers: bool = True,
        wait_for_more: bool = True,
        msecs: int = 0,
    ):
        flag = core.EventLoop.ProcessEventsFlag.AllEvents
        if not user_input:
            flag |= core.EventLoop.ProcessEventsFlag.ExcludeUserInputEvents
        if not socket_notifiers:
            flag |= core.EventLoop.ProcessEventsFlag.ExcludeSocketNotifiers
        if wait_for_more:  # doesnt seem to work? Could use this for sleep() otherwise.
            flag |= core.EventLoop.ProcessEventsFlag.WaitForMoreEvents
        cls.processEvents(flag, msecs)


class Application(ApplicationMixin, widgets.QApplication):
    """Manages the GUI application's control flow and main settings."""

    # def __init__(self, *args, **kwargs):
    #     super().__init__()


# Application.setStyle(widgets.Style())

if __name__ == "__main__":
    app = widgets.app()

    container = widgets.Widget()
    container.set_layout("horizontal")
    w = widgets.PlainTextEdit(parent=container)
    w2 = widgets.PlainTextEdit(parent=container)
    container.box.add(w)
    container.box.add(w2)
    container.show()
    # editor = widgeteditor.WidgetEditor(w)
    # editor.show()
    with app.debug_mode():
        app.sleep(1)
        app.exec()



from __future__ import annotations

from ctypes import POINTER, Structure, c_int, cast
from ctypes.wintypes import HWND, LPRECT, MSG, RECT, UINT
from enum import Enum
from sys import getwindowsversion

import win32con

from prettyqt import constants, core, gui, widgets
from prettyqt.utils.platforms.windows import misc, windoweffects


class PWINDOWPOS(Structure):
    _fields_ = [
        ("hWnd", HWND),
        ("hwndInsertAfter", HWND),
        ("x", c_int),
        ("y", c_int),
        ("cx", c_int),
        ("cy", c_int),
        ("flags", UINT),
    ]


class NCCALCSIZE_PARAMS(Structure):
    _fields_ = [("rgrc", RECT * 3), ("lppos", POINTER(PWINDOWPOS))]


LPNCCALCSIZE_PARAMS = POINTER(NCCALCSIZE_PARAMS)


def invert_color(color: str) -> str:
    inverted_color = ""
    for i in range(0, 5, 2):
        channel = int(color[i : i + 2], base=16)
        inverted_color += hex(round(channel / 6))[2:].upper().zfill(2)
    inverted_color += color[-2:]
    return inverted_color


def get_resize_border_thickness(h_wnd):
    window = gui.GuiApplication.find_window(h_wnd)
    if not window:
        return 0
    thickness = misc.get_border_thickness()
    return round(thickness * window.devicePixelRatio())


class CustomBase(widgets.Widget):
    def __init__(
        self, use_mica: bool = False, theme: str = "auto", color: str = "F0F0F0A0"
    ):
        """Customizable window without titlebar.

        Parameters
        ----------
        use_mica: 'bool
            Use mica or acrylic effect, (Mica Win11 only)
        theme: 'auto', 'dark', 'light'
            Use dark, light or system theme
        color: str
            Window background color
        """
        super().__init__()
        self.is_win11 = getwindowsversion().build >= 22000
        self.use_mica = use_mica if self.is_win11 else False
        match theme:
            case "auto":
                self.dark_mode = self.get_palette().is_dark()
            case "dark":
                self.dark_mode = True
            case "light":
                self.dark_mode = False
            case _:
                raise ValueError(theme)
        if len(color) != 8:
            raise ValueError("Wrong argument 'color': must be 8 hex symbols")
        self.acrylic_color = invert_color(color) if self.dark_mode else color
        self.effect_enabled = False
        self.win_effects = windoweffects.WindowsEffects()
        self.win_effects.add_window_animation(self.winId())
        self.setWindowFlags(constants.WindowType.FramelessWindowHint)
        self.win_effects.add_window_animation(self.winId())
        self.set_effect()
        if self.is_win11:
            self.win_effects.add_blur_behind_window(self.winId())
            self.win_effects.add_shadow_effect(self.winId())
        self.setStyleSheet("CustomBase{ background:transparent; }")

        self._effect_timer = core.Timer(self, interval=100, single_shot=True)
        self._effect_timer.timeout.connect(self.set_effect)

    def set_effect(self, enable: bool = True):
        if self.effect_enabled == enable:
            return
        self.effect_enabled = enable
        if enable and self.use_mica:
            self.win_effects.add_mica_effect(self.winId(), self.dark_mode)
        elif enable:
            self.win_effects.add_acrylic_effect(self.winId(), self.acrylic_color)
        else:
            self.win_effects.remove_background_effect(self.winId())
        self.update()

    def _temporary_disable_effect(self):
        self.set_effect(False)
        self._effect_timer.stop()
        self._effect_timer.start()

    def moveEvent(self, event):
        if self.is_win11 or not self._effect_timer:
            return super().moveEvent(event)
        self._temporary_disable_effect()

    def paintEvent(self, event):
        if self.effect_enabled:
            return super().paintEvent(event)
        with gui.Painter(self) as painter:
            painter.setOpacity(0.8)
            if self.dark_mode:
                painter.setBrush(constants.GlobalColor.black)
            else:
                painter.setBrush(constants.GlobalColor.white)
            painter.drawRect(self.rect())

    def nativeEvent(self, event_type, message):
        msg = MSG.from_address(int(message))
        if not msg.hWnd:
            return False, 0
        if msg.message == win32con.WM_NCCALCSIZE:
            if msg.wParam:
                rect = cast(msg.lParam, LPNCCALCSIZE_PARAMS).contents.rgrc[0]
            else:
                rect = cast(msg.lParam, LPRECT).contents

            is_max = misc.is_maximized(msg.hWnd)
            is_full = misc.is_full_screen(msg.hWnd)

            # Adjust the size of client rect
            if is_max and not is_full:
                thickness = misc.get_resize_border_thickness(msg.hWnd)
                rect.top += thickness
                rect.left += thickness
                rect.right -= thickness
                rect.bottom -= thickness

            # Handle the situation that an auto-hide taskbar is enabled
            if (is_max or is_full) and misc.Taskbar.is_auto_hide():
                match misc.Taskbar.get_position(msg.hWnd):
                    case misc.Taskbar.LEFT:
                        rect.top += misc.Taskbar.AUTO_HIDE_THICKNESS
                    case misc.Taskbar.BOTTOM:
                        rect.bottom -= misc.Taskbar.AUTO_HIDE_THICKNESS
                    case misc.Taskbar.LEFT:
                        rect.left += misc.Taskbar.AUTO_HIDE_THICKNESS
                    case misc.Taskbar.RIGHT:
                        rect.right -= misc.Taskbar.AUTO_HIDE_THICKNESS

            res = 0 if not msg.wParam else win32con.WVR_REDRAW
            return True, res

        return False, 0


class TitleBarButtonState(Enum):
    NORMAL = 0
    HOVER = 1
    PRESSED = 2


class TitleBarButton(widgets.ToolButton):
    def __init__(self, parent, dark_mode: bool):
        super().__init__(parent)
        if dark_mode:
            color = "FFFFFF"
            self._icon_color = constants.GlobalColor.white
        else:
            color = "000000"
            self._icon_color = constants.GlobalColor.black
        self.colors = "transparent", f"#20{color}", f"#40{color}"
        self._style = """
        border: none;
        margin: 0px;
        """
        self._state = TitleBarButtonState.NORMAL
        self.set_state(TitleBarButtonState.NORMAL)
        self.setFixedSize(46, 32)

    def get_state(self):
        return self._state

    def set_state(self, state):
        self._state = state
        self.setStyleSheet(
            f"background-color: {self.colors[state.value]};\n{self._style}"
        )

    def enterEvent(self, e):
        self.set_state(TitleBarButtonState.HOVER)
        super().enterEvent(e)

    def leaveEvent(self, e):
        self.set_state(TitleBarButtonState.NORMAL)
        super().leaveEvent(e)

    def mousePressEvent(self, e):
        if e.button() == constants.MouseButton.LeftButton:
            self.set_state(TitleBarButtonState.PRESSED)
        super().mousePressEvent(e)

    def mouseReleaseEvent(self, e):
        self.set_state(TitleBarButtonState.HOVER)
        super().mouseReleaseEvent(e)


class MinimizeButton(TitleBarButton):
    def paintEvent(self, event):
        super().paintEvent(event)
        with gui.Painter(self) as painter:
            pen = gui.QPen(self._icon_color)
            pen.setCosmetic(True)
            painter.setPen(pen)
            painter.drawLine(18, 16, 28, 16)


class MaximizeButton(TitleBarButton):
    def __init__(self, parent, dark_mode):
        super().__init__(parent, dark_mode)
        self.is_max = False

    def paintEvent(self, event):
        super().paintEvent(event)
        with gui.Painter(self) as painter:
            pen = gui.QPen(self._icon_color)
            pen.setCosmetic(True)
            painter.setPen(pen)

            r = self.devicePixelRatioF()
            painter.scale(1 / r, 1 / r)
            if not self.is_max:
                painter.drawRect(int(18 * r), int(11 * r), int(10 * r), int(10 * r))
            else:
                r_18, r_8 = int(18 * r), int(8 * r)
                painter.drawRect(r_18, int(13 * r), r_8, r_8)
                x0 = r_18 + int(2 * r)
                y0 = 13 * r
                dw = int(2 * r)
                path = gui.PainterPath(core.QPointF(x0, y0))
                path.lineTo(x0, y0 - dw)
                path.lineTo(x0 + 8 * r, y0 - dw)
                path.lineTo(x0 + 8 * r, y0 - dw + 8 * r)
                path.lineTo(x0 + 8 * r - dw, y0 - dw + 8 * r)
                painter.drawPath(path)


class CloseButton(TitleBarButton):
    def __init__(self, parent, dark_mode):
        super().__init__(parent, dark_mode)
        self._dark_mode = dark_mode
        self.colors = "transparent", "#C42B1C", "#C83C30"
        self.set_state(TitleBarButtonState.NORMAL)
        self._white_icon = gui.Icon(r"prettyqt\resources\close_white.svg")
        if self._dark_mode:
            self.setIcon(self._white_icon)
        else:
            self._black_icon = gui.Icon(r"prettyqt\resources\close_black.svg")
            self.setIcon(self._black_icon)
        self.setIconSize(core.QSize(46, 32))

    def enterEvent(self, event):
        if not self._dark_mode:
            self.setIcon(self._white_icon)
        super().enterEvent(event)

    def leaveEvent(self, event):
        if not self._dark_mode:
            self.setIcon(self._black_icon)
        super().leaveEvent(event)


class TitleBar(widgets.Widget):
    def __init__(self, parent, dark_mode):
        super().__init__(parent)
        self.setFixedHeight(32)

        self.icon = widgets.Label(self)
        self.title = widgets.Label(self)
        self.min_btn = MinimizeButton(self, dark_mode)
        self.max_btn = MaximizeButton(self, dark_mode)
        self.close_btn = CloseButton(self, dark_mode)
        self.h_box_layout = widgets.HBoxLayout(self, spacing=0, margin=0)

        if dark_mode:
            self.title.setStyleSheet("color: white")
        self.icon.setFixedSize(10, 16)

        self.h_box_layout.addWidget(self.icon)
        self.h_box_layout.addWidget(self.title)
        self.h_box_layout.addWidget(self.min_btn)
        self.h_box_layout.addWidget(self.max_btn)
        self.h_box_layout.addWidget(self.close_btn)

        self.min_btn.clicked.connect(self.window().showMinimized)
        self.max_btn.clicked.connect(self.__toggle_max_state)
        self.close_btn.clicked.connect(self.window().close)

    def __toggle_max_state(self):
        is_max = self.window().isMaximized()
        self.max_btn.is_max = not is_max
        if is_max:
            self.window().showNormal()
        else:
            self.window().showMaximized()

    def mouseDoubleClickEvent(self, event):
        if event.button() == constants.MouseButton.LeftButton:
            self.__toggle_max_state()

    def mouseMoveEvent(self, event):
        if not event.pos().x() < self.width() - 46 * 3:
            return
        win_id = self.window().winId()
        misc.start_system_move(win_id)


class FramelessWindow(CustomBase):
    BORDER_WIDTH = 4

    def __init__(self, use_mica=True, theme="auto", color="F2F2F299"):
        """Customizable window with custom titlebar.

        Parameters
        ----------
        use_mica: 'false', 'true', 'if available'
            Use mica or acrylic effect,
            'if available' mode will select according to the current OS
        theme: 'auto', 'dark', 'light'
            Use dark, light or system theme
        color: str
            Window background color
        """
        self._max_btn_hovered = False
        self.title_bar = None
        super().__init__(use_mica, theme, color)
        self.title_bar = TitleBar(self, self.dark_mode)
        self.setContentsMargins(0, 32, 0, 0)

    def setWindowTitle(self, title):
        self.title_bar.title.setText(title)
        super().setWindowTitle(title)

    def setWindowIcon(self, icon):
        self.title_bar.icon.setFixedWidth(32)
        self.title_bar.icon.setAlignment(constants.AlignmentFlag.AlignHCenter)
        self.title_bar.icon.setPixmap(icon.pixmap(16, 16))
        super().setWindowIcon(icon)

    def resizeEvent(self, event):
        if not self.title_bar:  # if not initialized
            return
        self.title_bar.setFixedWidth(self.width())
        if not self.use_mica:
            self._temporary_disable_effect()

    def nativeEvent(self, event_type, message):
        msg = MSG.from_address(int(message))
        if not msg.hWnd:
            return False, 0
        if msg.message == win32con.WM_NCHITTEST:
            pos = gui.Cursor.pos()
            x = pos.x() - self.x()
            y = pos.y() - self.y()
            if (
                self.is_win11
                and self.title_bar.childAt(pos - self.geometry().topLeft())
                is self.title_bar.max_btn
            ):
                self._max_btn_hovered = True
                self.title_bar.max_btn.set_state(TitleBarButtonState.HOVER)
                return True, win32con.HTMAXBUTTON
            lx = x < self.BORDER_WIDTH
            rx = x > self.width() - self.BORDER_WIDTH
            ty = y < self.BORDER_WIDTH
            by = y > self.height() - self.BORDER_WIDTH
            if rx and by:
                return True, win32con.HTBOTTOMRIGHT
            elif rx and ty:
                return True, win32con.HTTOPRIGHT
            elif lx and by:
                return True, win32con.HTBOTTOMLEFT
            elif lx and ty:
                return True, win32con.HTTOPLEFT
            elif rx:
                return True, win32con.HTRIGHT
            elif by:
                return True, win32con.HTBOTTOM
            elif lx:
                return True, win32con.HTLEFT
            elif ty:
                return True, win32con.HTTOP
        elif self.is_win11 and self._max_btn_hovered:
            max_btn_state = self.title_bar.max_btn.get_state()
            match msg.message:
                case win32con.WM_NCLBUTTONDOWN:
                    self.title_bar.max_btn.set_state(TitleBarButtonState.PRESSED)
                    return True, 0
                case win32con.WM_NCLBUTTONUP | win32con.WM_NCRBUTTONUP:
                    self.title_bar.max_btn.click()
                case 0x2A2 | win32con.WM_MOUSELEAVE if max_btn_state != 0:
                    self._max_btn_hovered = False
                    self.title_bar.max_btn.set_state(TitleBarButtonState.NORMAL)

        return super().nativeEvent(event_type, message)


if __name__ == "__main__":
    app = widgets.app()
    app.set_style("Fusion")
    win = FramelessWindow()
    win.setWindowTitle("Thsi is  a tesxt")
    win.show()
    app.exec()



from __future__ import annotations

import logging

from prettyqt import core
from prettyqt.utils import node


logger = logging.getLogger(__name__)


class TreeItem(node.Node):
    """Tree node class used for Qt TreeModels."""

    # __slots__ = ("has_children", "children_fetched")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.has_children = True
        self.children_fetched = False


class TreeModel(core.AbstractItemModel):
    """TreeModel which is used as a base for most TreeModels implemented.

    Tree gets dynamically populated whenever the the indexes get "expanded".

    basically, only 2 methods need to be defined in order to build the tree:

    def _fetch_object_children(self, TreeModel.TreeItem) -> list[TreeModel.TreeItem]:
        ...

    def _has_children(self, TreeModel.TreeItem) -> bool:
        ...

    Subclasses can override TreeModel.TreeItem if a child TreeItem class should be used
    for the root.

    """

    TreeItem = TreeItem

    def __init__(self, obj=None, show_root: bool = True, **kwargs):
        super().__init__(**kwargs)
        self._root_item = self.TreeItem(obj=obj)
        self._show_root = show_root
        self.set_root_item(obj)

    @property
    def show_root(self) -> bool:
        """Return True if the inspected node is visible.

        In that case an invisible root node has been added.
        """
        return self._show_root

    def set_root_item(self, obj):
        if self._show_root:
            self._root_item = self.TreeItem(obj=None)
            self._root_item.children_fetched = True
            self.inspected_item = self.TreeItem(obj=obj)
            self._root_item.append_child(self.inspected_item)
            # root_index = self.index(0, 0)
            # self.fetchMore(self.index(0, 0, root_index))
        else:
            # The root itself will be invisible
            self._root_item = self.TreeItem(obj=obj)
            self.inspected_item = self._root_item
            root_index = self.index(0, 0)
            self.fetchMore(root_index)

    @property
    def root_item(self) -> TreeModel.TreeItem:
        return self._root_item

    def data_by_index(self, index: core.ModelIndex) -> TreeModel.TreeItem:
        return index.internalPointer() if index.isValid() else self.root_item

    def index_for_item(
        self, item: TreeModel.TreeItem, column: int = 0
    ) -> core.ModelIndex:
        if item == self.root_item:
            return core.ModelIndex()
        parent = item.parent_item or self.root_item
        if (not parent) or (len(parent.children) == 0):
            return None
        row = parent.children.index(item)
        return self.createIndex(row, column, item)

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        if not self.hasIndex(row, column, parent):
            return core.ModelIndex()
        parent_item = self.data_by_index(parent)
        child_item = parent_item.children[row]
        return self.createIndex(row, column, child_item)

    def parent(self, index: core.ModelIndex | None = None) -> core.ModelIndex:
        # hacky way to let the case without any arguments get through.
        # not really nice, a proper dispatch library would be better.
        # functools.singledispatchmethod doesnt work here.
        index = index or core.ModelIndex()
        if not index.isValid():
            return core.ModelIndex()

        child_item = self.data_by_index(index)
        parent_item = child_item.parent_item

        if parent_item is None or parent_item == self.root_item:
            return core.ModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return 0 if parent.column() > 0 else len(self.data_by_index(parent).children)

    def hasChildren(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        treeitem = self.data_by_index(parent)
        if self._show_root and treeitem == self._root_item:
            return True
        return self._has_children(treeitem)

    def canFetchMore(self, parent: core.ModelIndex | None = None):
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        else:
            return not self.data_by_index(parent).children_fetched

    def fetchMore(self, parent: core.ModelIndex | None = None):
        """Fetch the children given the model index of a parent node.

        Adds the children to the parent.
        """
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return

        parent_item = self.data_by_index(parent)
        if parent_item.children_fetched:
            return

        tree_items = self._fetch_object_children(parent_item)

        with self.insert_rows(0, len(tree_items) - 1, parent):
            for tree_item in tree_items:
                parent_item.append_child(tree_item)
            parent_item.children_fetched = True

    def _fetch_object_children(self, treeitem) -> list[TreeModel.TreeItem]:
        return treeitem.children

    def _has_children(self, treeitem) -> bool:
        return treeitem.has_children


if __name__ == "__main__":
    model = TreeModel("test")
    model.parent()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtQml
from prettyqt.utils import bidict, datatypes


COMPILATION_MODES = bidict(
    prefer_synchronous=QtQml.QQmlComponent.CompilationMode.PreferSynchronous,
    asynchronous=QtQml.QQmlComponent.CompilationMode.Asynchronous,
)

STATUS = bidict(
    null=QtQml.QQmlComponent.Status.Null,
    ready=QtQml.QQmlComponent.Status.Ready,
    loading=QtQml.QQmlComponent.Status.Loading,
    error=QtQml.QQmlComponent.Status.Error,
)


class QmlComponent(core.ObjectMixin, QtQml.QQmlComponent):
    """Encapsulates a QML component definition."""

    def get_status(self) -> str:
        return STATUS.inverse[self.status()]

    def get_url(self) -> core.Url:
        return core.Url(self.url())

    def load_url(self, url: datatypes.UrlType, mode: str):
        if isinstance(url, str):
            url = core.Url.from_user_input(url)
        self.loadUrl(url, COMPILATION_MODES[mode])



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtLocation
from prettyqt.utils import bidict


try:  # PySide2 5.15 doesnt include these
    VISIBILITY = bidict(
        unspecified=QtLocation.QLocation.UnspecifiedVisibility,  # type: ignore
        device=QtLocation.QLocation.DeviceVisibility,  # type: ignore
        private=QtLocation.QLocation.PrivateVisibility,  # type: ignore
        public=QtLocation.QLocation.PublicVisibility,  # type: ignore
    )
except AttributeError:
    VISIBILITY = bidict()


VisibilityStr = Literal["unspecified", "device", "private", "public"]



from __future__ import annotations

from prettyqt import gui
from prettyqt.qt import QtWidgets


class WidgetAction(gui.ActionMixin, QtWidgets.QWidgetAction):
    pass


if __name__ == "__main__":
    w = WidgetAction()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtPositioning
from prettyqt.utils import bidict, get_repr


ERROR = bidict(
    access_error=QtPositioning.QGeoSatelliteInfoSource.Error.AccessError,
    closed_error=QtPositioning.QGeoSatelliteInfoSource.Error.ClosedError,
    none=QtPositioning.QGeoSatelliteInfoSource.Error.NoError,
    unknown_source=QtPositioning.QGeoSatelliteInfoSource.Error.UnknownSourceError,
)

ErrorStr = Literal["access_error", "closed_error", "none", "unknown_source"]


class GeoSatelliteInfoSource(core.ObjectMixin, QtPositioning.QGeoSatelliteInfoSource):
    def __str__(self):
        return self.sourceName()

    def __repr__(self):
        return get_repr(self)

    def get_error(self) -> ErrorStr:
        """Return error type.

        Returns:
            error type
        """
        return ERROR.inverse[self.error()]



from __future__ import annotations

from typing import Literal

from prettyqt import constants, gui, widgets
from prettyqt.utils import bidict


ShapeModeStr = Literal["mask", "bounding_rect", "heuristic_mask"]

SHAPE_MODE: bidict[ShapeModeStr, widgets.QGraphicsPixmapItem.ShapeMode] = bidict(
    mask=widgets.QGraphicsPixmapItem.ShapeMode.MaskShape,
    bounding_rect=widgets.QGraphicsPixmapItem.ShapeMode.BoundingRectShape,
    heuristic_mask=widgets.QGraphicsPixmapItem.ShapeMode.HeuristicMaskShape,
)


class GraphicsPixmapItem(widgets.GraphicsItemMixin, widgets.QGraphicsPixmapItem):
    """Pixmap item that you can add to a QGraphicsScene."""

    def set_transformation_mode(
        self, mode: constants.TransformationModeStr | constants.TransformationMode
    ):
        """Set transformation mode.

        Args:
            mode: transformation mode to use
        """
        self.setTransformationMode(constants.TRANSFORMATION_MODE.get_enum_value(mode))

    def get_transformation_mode(self) -> constants.TransformationModeStr:
        """Return current transformation mode.

        Returns:
            transformation mode
        """
        return constants.TRANSFORMATION_MODE.inverse[self.transformationMode()]

    def set_shape_mode(self, mode: ShapeModeStr | widgets.QGraphicsPixmapItem.ShapeMode):
        """Set shape mode.

        Args:
            mode: shape mode to use
        """
        self.setShapeMode(SHAPE_MODE.get_enum_value(mode))

    def get_shape_mode(self) -> ShapeModeStr:
        """Return current shape mode.

        Returns:
            shape mode
        """
        return SHAPE_MODE.inverse[self.shapeMode()]

    def get_pixmap(self) -> gui.Pixmap | None:
        pix = self.pixmap()
        return None if pix.isNull() else gui.Pixmap(pix)



from __future__ import annotations

from collections.abc import Iterator
from re import Pattern

from prettyqt import core, gui


class SyntaxHighlighterMixin(core.ObjectMixin):
    RULES: list = []

    def __init__(self, parent: core.QObject | None = None):
        super().__init__(parent)  # type: ignore

    def get_current_block(self) -> gui.TextBlock:
        return gui.TextBlock(self.currentBlock())

    def get_format(self, position: int) -> gui.TextCharFormat:
        return gui.TextCharFormat(self.format(position))

    @classmethod
    def iter_rules(cls) -> Iterator[tuple[Pattern, int, gui.TextCharFormat]]:
        for rule in cls.RULES:
            if isinstance(rule.compiled, list):
                for i in rule.compiled:
                    yield (i, rule.nth, rule.fmt)
            else:
                yield (rule.compiled, rule.nth, rule.fmt)

    def highlightBlock(self, text: str):
        """Apply syntax highlighting to the given block of text."""
        # Do other syntax formatting
        for expression, nth, fmt in self.iter_rules():
            for match in expression.finditer(text):
                span = match.span(nth)
                self.setFormat(span[0], span[1] - span[0], fmt)


class SyntaxHighlighter(SyntaxHighlighterMixin, gui.QSyntaxHighlighter):
    pass



from __future__ import annotations

from typing import Literal

from typing_extensions import Self

from prettyqt.qt import QtCore
from prettyqt.utils import bidict, get_repr


StringFormatStr = Literal["with_braces", "without_braces", "id_128"]

STRING_FORMATS: bidict[StringFormatStr, QtCore.QUuid.StringFormat] = bidict(
    with_braces=QtCore.QUuid.StringFormat.WithBraces,
    without_braces=QtCore.QUuid.StringFormat.WithoutBraces,
    id_128=QtCore.QUuid.StringFormat.Id128,
)

VariantStr = Literal["unknown", "ncs", "dce", "microsoft", "reserved"]

VARIANTS: bidict[VariantStr, QtCore.QUuid.Variant] = bidict(
    unknown=QtCore.QUuid.Variant.VarUnknown,
    ncs=QtCore.QUuid.Variant.NCS,
    dce=QtCore.QUuid.Variant.DCE,
    microsoft=QtCore.QUuid.Variant.Microsoft,
    reserved=QtCore.QUuid.Variant.Reserved,
)

VersionStr = Literal["unknown", "time", "embedded_posix", "name", "random", "sha1"]

VERSION: bidict[VersionStr, QtCore.QUuid.Version] = bidict(
    unknown=QtCore.QUuid.Version.VerUnknown,
    time=QtCore.QUuid.Version.Time,
    embedded_posix=QtCore.QUuid.Version.EmbeddedPOSIX,
    name=QtCore.QUuid.Version.Name,
    # md5=QtCore.QUuid.Version.Md5,
    random=QtCore.QUuid.Version.Random,
    sha1=QtCore.QUuid.Version.Sha1,
)


class UuidMixin:
    def __repr__(self):
        return get_repr(self, self.toString())

    def __str__(self):
        return self.toString()

    def __bool__(self):
        return not self.isNull()

    def __reduce__(self):
        return type(self), (self.toString(),)

    def __format__(self, format_spec: StringFormatStr):
        return self.to_string(format_spec)

    def get_variant(self) -> VariantStr:
        return VARIANTS.inverse[self.variant()]

    def get_version(self) -> VersionStr:
        return VERSION.inverse[self.version()]

    @classmethod
    def create_uuid(cls) -> Self:
        # workaround for PySide2, not able to clone in ctor
        return cls(cls.createUuid().toString())

    def to_string(
        self, fmt: StringFormatStr | QtCore.QUuid.StringFormat = "with_braces"
    ) -> str:
        """Return string representation of the Uuid.

        Allowed values are "with_braces", "without_braces", "id_128"

        Args:
            fmt: Uuid format to use
        """
        return self.toString(STRING_FORMATS.get_enum_value(fmt))


class Uuid(UuidMixin, QtCore.QUuid):
    """Stores a Universally Unique Identifier (UUID)."""


if __name__ == "__main__":
    address = Uuid.create_uuid()
    print(repr(address))



from __future__ import annotations

# from collections.abc import Sequence
import contextlib
import enum

from importlib import metadata
import logging
import pkgutil

from packaging.markers import Marker
from packaging.requirements import InvalidRequirement, Requirement
from typing_extensions import Self

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


def load_dist_info(name: str) -> metadata.Distribution | None:
    try:
        return metadata.distribution(name)
    except metadata.PackageNotFoundError:
        return None


def list_system_modules() -> list[metadata.Distribution]:
    modules = (test.name for test in pkgutil.iter_modules())
    distributions = (load_dist_info(i) for i in modules)
    return [i for i in distributions if i is not None]


def list_package_requirements(package_name: str) -> list[metadata.Distribution]:
    dist = metadata.distribution(package_name)
    try:
        modules = {Requirement(i).name for i in dist.requires} if dist.requires else set()
    except ValueError as e:
        logger.error(f"{e} for {dist.name}")
        return []
    return [dist for i in modules if (dist := load_dist_info(i)) is not None]


class DistTreeItem(itemmodels.ColumnItemModel.TreeItem):
    # __slots__ = ("requires", "metadata", "version", "markers")

    def __init__(
        self,
        obj,
        parent: DistTreeItem | None = None,
    ):
        super().__init__(obj, parent=parent)
        self.requires = {}
        self.markers = {}
        if obj is not None and obj.requires is not None:
            for i in obj.requires:
                with contextlib.suppress(InvalidRequirement):
                    req = Requirement(i)
                    self.requires[req.name] = req
                    if ";" in i:
                        extras_str = i.split(";", maxsplit=1)[-1]
                        self.markers[req.name] = Marker(extras_str)
        # Cache the data to avoid excessive IO (obj.metadata reads file)
        self.metadata = None if obj is None else obj.metadata
        self.version = None if obj is None else obj.version


class DistributionColumn(itemmodels.ColumnItem):
    def get_data(self, item: ImportlibTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case ImportlibTreeModel.Roles.DistributionRole:
                return item.obj


class NameColumn(DistributionColumn):
    name = "Name"
    doc = "Package name"

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["Name"] or ""
            case _:
                return super().get_data(item, role)


class VersionColumn(DistributionColumn):
    name = "Version"
    doc = "Version number."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.version or ""
            case _:
                return super().get_data(item, role)


class ConstraintsColumn(DistributionColumn):
    name = "Constraints"
    doc = "Constraints."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return next(
                    (
                        str(requirement.specifier)
                        for requirement in item.parent_item.requires.values()
                        if item.metadata["name"] == requirement.name
                    ),
                    "",
                )
            case _:
                return super().get_data(item, role)


class MarkerColumn(DistributionColumn):
    name = "Extra"
    doc = "Extra."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return next(
                    (
                        str(marker)
                        for name, marker in item.parent_item.markers.items()
                        if item.metadata["name"] == name
                    ),
                    "",
                )
            case _:
                return super().get_data(item, role)


class SummaryColumn(DistributionColumn):
    name = "Summary"
    doc = "Module description."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["Summary"] or ""
            case _:
                return super().get_data(item, role)


class HomepageColumn(DistributionColumn):
    name = "Homepage"
    doc = "URL of the homepage."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["Home-Page"] or ""
            case _:
                return super().get_data(item, role)


class AuthorColumn(DistributionColumn):
    name = "Author"
    doc = "Author name."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["Author"] or ""
            case _:
                return super().get_data(item, role)


class LicenseColumn(DistributionColumn):
    name = "License"
    doc = "License name."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["License"] or ""
            case _:
                return super().get_data(item, role)


class ImportlibTreeModel(itemmodels.ColumnItemModel):
    """Model showing the dependency tree of a distribution."""

    @core.Enum
    class Roles(enum.IntEnum):
        DistributionRole = constants.USER_ROLE + 43255

    SUPPORTS = metadata.Distribution | str
    TreeItem = DistTreeItem
    IS_RECURSIVE = True
    COLUMNS = [
        NameColumn,
        VersionColumn,
        ConstraintsColumn,
        MarkerColumn,
        SummaryColumn,
        HomepageColumn,
        AuthorColumn,
        LicenseColumn,
    ]

    def __init__(
        self,
        obj: metadata.Distribution | str,
        show_root: bool = False,
        parent: core.QObject | None = None,
    ):
        if isinstance(obj, str):
            obj = metadata.distribution(obj)
        super().__init__(
            obj=obj, columns=self.COLUMNS, parent=parent, show_root=show_root
        )

    @classmethod
    def setup_example(cls):
        from prettyqt import widgets

        model = cls("prettyqt")
        table = widgets.TreeView(word_wrap=False)
        table.set_model(model)
        table.set_delegate("render_link", column=5)
        table.expand_all(depth=4)
        return table

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, metadata.Distribution)

    @classmethod
    def from_system(cls, parent: core.QObject | None = None) -> Self:
        distributions = list_system_modules()
        return cls(distributions, parent)

    @classmethod
    def from_package(cls, package_name: str, parent: core.QObject | None = None) -> Self:
        distributions = list_package_requirements(package_name)
        return cls(distributions, parent)

    def _has_children(self, item: ImportlibTreeModel.TreeItem) -> bool:
        return bool(item.requires)

    def _fetch_object_children(self, item: DistTreeItem) -> list[DistTreeItem]:
        return [
            DistTreeItem(obj=dist, parent=item)
            for dist in list_package_requirements(item.metadata["Name"])
        ]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    model = ImportlibTreeModel("prettyqt")
    table = widgets.TreeView(word_wrap=False)
    # table.setSortingEnabled(True)
    table.set_model(model)
    table.set_delegate("render_link", column=5)
    table.expand_all(depth=4)
    table.show()
    app.exec()



from __future__ import annotations

from prettyqt.qt import QtCore
from prettyqt.utils import datatypes


class SerializeMixin:
    def __getstate__(self):
        ba = self.create_bytearray()
        return ba.data()

    def __setstate__(self, ba):
        self.write_bytearray(ba)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __bytes__(self):
        return self.__getstate__()

    def create_bytearray(self) -> QtCore.QByteArray:
        ba = QtCore.QByteArray()
        stream = QtCore.QDataStream(ba, QtCore.QIODeviceBase.OpenModeFlag.WriteOnly)
        stream << self
        return ba

    def write_bytearray(self, ba: datatypes.ByteArrayType):
        if isinstance(ba, str):
            ba = ba.encode()
        if not isinstance(ba, QtCore.QByteArray):
            ba = QtCore.QByteArray(ba)
        stream = QtCore.QDataStream(ba, QtCore.QIODeviceBase.OpenModeFlag.ReadOnly)
        stream >> self

    def copy_data_to(self, dest: datatypes.QtSerializableType):
        ba = self.create_bytearray()
        stream = QtCore.QDataStream(ba, QtCore.QIODeviceBase.OpenModeFlag.ReadOnly)
        stream >> dest



from __future__ import annotations

import datetime

from prettyqt import charts
from prettyqt.utils import datatypes


class DateTimeAxis(charts.AbstractAxisMixin, charts.QDateTimeAxis):
    def set_min(self, minimum: datatypes.DateTimeType):
        """Set minimum value for axis."""
        minimum = datatypes.to_datetime(minimum)
        self.setMin(minimum)

    def get_min(self) -> datetime.datetime:
        return self.min().toPython()

    def set_max(self, maximum: datatypes.DateTimeType):
        """Set maximum value for axis."""
        maximum = datatypes.to_datetime(maximum)
        self.setMax(maximum)

    def get_max(self) -> datetime.datetime:
        return self.max().toPython()

    def set_range(self, minimum: datatypes.DateTimeType, maximum: datatypes.DateTimeType):
        """Set value range of datetime axis."""
        self.setRange(datatypes.to_datetime(minimum), datatypes.to_datetime(maximum))



from __future__ import annotations

from prettyqt import core, qml
from prettyqt.utils import bidict


EXTENSIONS = bidict(
    translation=qml.QJSEngine.Extension.TranslationExtension,
    console=qml.QJSEngine.Extension.ConsoleExtension,
    garbage_collection=qml.QJSEngine.Extension.GarbageCollectionExtension,
    all=qml.QJSEngine.Extension.AllExtensions,
)


class JSEngineMixin(core.ObjectMixin):
    def install_extensions(self, extension: str, obj: qml.QJSValue | None = None):
        if obj is None:
            obj = qml.QJSValue()
        self.installExtensions(EXTENSIONS[extension], obj)

    def new_array(self, length: int = 0) -> qml.JSValue:
        return qml.JSValue(self.newArray(length))

    def eval(self, program: str) -> qml.JSValue:
        result = self.evaluate(program)
        return qml.JSValue(result)  # type: ignore


class JSEngine(JSEngineMixin, qml.QJSEngine):
    """Environment for evaluating JavaScript code."""


if __name__ == "__main__":
    app = core.app()
    engine = JSEngine()
    fun = engine.eval("(function(a, b) { return a + b; })")
    result = fun(1, 2)
    print(result)
